# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the request handlers of an-website."""

from __future__ import annotations

import socket
from datetime import datetime
from urllib.parse import quote_from_bytes

from html5lib import HTMLParser
from time_machine import travel
from tornado.simple_httpclient import SimpleAsyncHTTPClient

from an_website.utils.options import COLOUR_SCHEMES

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_url_query,
    assert_valid_dynload_response,
    assert_valid_html_response,
    assert_valid_json_response,
    assert_valid_redirect,
    assert_valid_response,
    assert_valid_rss_response,
    assert_valid_yaml_response,
    check_html_page,
    fetch,
    hill_valley,
)
from .test_settings import parse_cookie


async def test_json_apis(fetch: FetchCallable) -> None:  # noqa: F811
    """Check whether the APIs return valid JSON."""
    json_apis = (
        "/api/betriebszeit",
        "/api/discord",
        "/api/discord/367648314184826880",
        "/api/discord",
        "/api/endpunkte",
        "/api/hangman-loeser",
        "/api/ip",
        "/api/ping",
        "/api/version",
        "/api/vertauschte-woerter",
        "/api/waehrungs-rechner",
        "/api/wortspiel-helfer",
        # "/api/zitate/1-1",  # gets tested with quotes
    )
    for api in json_apis:
        json_resp = assert_valid_json_response(
            await fetch(api, headers={"Accept": "application/json"})
        )
        yaml_resp = assert_valid_yaml_response(
            await fetch(api, headers={"Accept": "application/yaml"})
        )
        if api != "/api/betriebszeit":
            assert json_resp == yaml_resp
        assert not assert_valid_response(
            await fetch(
                api, method="HEAD", headers={"Accept": "application/json"}
            ),
            "application/json",
        ).body
        assert not assert_valid_response(
            await fetch(
                api, method="HEAD", headers={"Accept": "application/yaml"}
            ),
            "application/yaml",
        ).body


async def test_invalid_utf8(fetch: FetchCallable) -> None:  # noqa: F811
    """Check that requests with invalid utf-8 work correctly."""
    replacement = "\uFFFD"
    for umlaut in "äöüÄÖÜ":
        latin1 = quote_from_bytes(umlaut.encode("latin1"))
        await assert_valid_redirect(fetch, f"/{latin1}", "/", {307})

        assert_valid_response(
            await fetch(f"/{latin1 * 100}"), "text/html;charset=utf-8", {404}
        )

        response = assert_valid_json_response(
            await fetch(f"/api/wortspiel-helfer?word={latin1}"), {200}
        )
        assert response["word"] == replacement

        response = assert_valid_json_response(
            await fetch(f"/api/hangman-loeser?input={latin1}"),
            {200},
        )
        assert response["input"] == replacement


async def test_not_found_handler(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the NotFoundHandler works."""
    assert_valid_html_response(await fetch("/qwertzuiop"), {404})
    assert_valid_html_response(
        await fetch(
            "/https:/github.com/asozialesnetzwerk/vertauschtewoerterplugin"
        ),
        {404},
    )

    await assert_valid_redirect(fetch, "/services.html", "/services", {308})
    await assert_valid_redirect(fetch, "/services/", "/services", {308})
    await assert_valid_redirect(fetch, "/services/////////", "/services", {308})

    await assert_valid_redirect(fetch, "/serices/index.htm", "/serices", {308})
    await assert_valid_redirect(fetch, "/serices", "/services", {307})

    await assert_valid_redirect(fetch, "/serwizes", "/services", {307})
    await assert_valid_redirect(fetch, "/service?x=y", "/services?x=y", {307})
    await assert_valid_redirect(fetch, "/servces?x=y", "/services?x=y", {307})

    await assert_valid_redirect(
        fetch,
        "/vertauschtewörter",
        "/vertauschte-woerter",
        {307},
        httpclient=SimpleAsyncHTTPClient(),
    )

    await assert_valid_redirect(
        fetch,
        "/vertauschtew%C3%B6rter",
        "/vertauschte-woerter",
        {307},
        httpclient=SimpleAsyncHTTPClient(),
    )
    await assert_valid_redirect(
        fetch,
        "/vertauschte-w%F6rter",
        "/vertauschte-woerter",
        {307},
        httpclient=SimpleAsyncHTTPClient(),
    )

    await assert_valid_redirect(fetch, "/a?x=y", "/?x=y", {307})

    await assert_valid_redirect(fetch, "/index.php", "/", {307})
    assert_valid_html_response(
        await fetch(
            "/index.php", method="POST", allow_nonstandard_methods=True
        ),
        {404},
    )

    assert_valid_html_response(await fetch("/wp-login.php"), {469})


async def test_page_crawling(
    fetch: FetchCallable,  # noqa: F811
    http_server_port: tuple[socket.socket, int],
) -> None:
    """Test most of the request handlers with crawling."""
    urls: set[str] = set()
    await check_html_page(fetch, "/", recursive=6, checked_urls=urls)
    for url in urls:
        assert_url_query(url, theme=None, no_3rd_party=None, dynload=None)

    urls_theme: set[str] = set()
    await check_html_page(
        fetch,
        "/?theme=pink",
        recursive=6,
        checked_urls=urls_theme,
    )
    for url in urls_theme:
        assert_url_query(url, theme="pink", no_3rd_party=None, dynload=None)

    urls_3rd_party: set[str] = set()
    await check_html_page(
        fetch,
        "/?no_3rd_party=sure",
        recursive=6,
        checked_urls=urls_3rd_party,
    )
    for url in urls_3rd_party:
        assert_url_query(url, theme=None, no_3rd_party="sure", dynload=None)

    urls_dynload: set[str] = set()
    await check_html_page(
        fetch,
        "/?dynload=sure",
        recursive=6,
        checked_urls=urls_dynload,
    )
    for url in urls_dynload:
        assert_url_query(url, theme=None, no_3rd_party=None, dynload="sure")


async def test_request_handlers0(
    fetch: FetchCallable,  # noqa: F811
    http_server_port: tuple[socket.socket, int],
) -> None:
    """Test more request handler stuff."""
    response = await fetch("/", headers={"Host": "example.org"})
    assert response.headers["Onion-Location"] == "http://test.onion/"
    assert b"http://test.onion" in response.body
    assert_valid_html_response(response, effective_url="http://example.org")

    response = await fetch("/", headers={"Host": "test.onion"})
    assert "Onion-Location" not in response.headers
    assert_valid_html_response(response, effective_url="http://test.onion")

    assert_valid_response(
        await fetch("/", headers={"Accept": "text/spam"}), None, {406}
    )

    await assert_valid_redirect(
        fetch,
        "/redirect?to=/endpunkte?x=y",
        "/endpunkte?x=y",
    )

    await assert_valid_redirect(
        fetch,
        "/redirect?theme=blue&x=y&to=/betriebszeit",
        "/betriebszeit?theme=blue",
    )
    await assert_valid_redirect(fetch, "/redirect?to=", "/", method="HEAD")
    await assert_valid_redirect(
        fetch,
        "/redirect?to=/",
        "/",
    )
    # no "to" query param, so redirect to main page
    await assert_valid_redirect(
        fetch,
        "/redirect?theme=blue",
        "/?theme=blue",
    )
    # no redirect
    assert "https://evil.com" in assert_valid_html_response(
        await fetch(
            "/redirect?theme=blue&from=/&to=https://evil.com",
        )
    ).body.decode("UTF-8")
    await fetch(
        "/redirect?theme=blue&from=/&to=https://evil.com",
        method="HEAD",
    )

    assert (
        assert_valid_response(
            await fetch("/", method="HEAD"), "text/html;charset=utf-8"
        ).body
        == b""
    )


@travel(datetime(2000 + 1, 2, 3, 4, 5, 6, 7, hill_valley), tick=False)
async def test_request_handlers1(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the request handlers return 200 codes."""
    for theme in ("default", "random", "fun", "christmas"):
        response = assert_valid_html_response(await fetch(f"/?theme={theme}"))
        assert response.code == 200
        body = response.body.decode("UTF-8")
        if theme != "default":
            assert f"theme={theme}" in body
        if not theme.startswith("random"):
            assert f"/{theme}.css" in body
        assert "🦘" in body[body.find("Mit Liebe gebacken") :]
        assert "🦘" in body

    for val in ("img", "glyf_colr0", "glyf_colr1"):
        response = await check_html_page(fetch, f"/?openmoji={val}")
        assert response.code == 200
        body = response.body.decode("UTF-8")
        assert f'"{val}"' in body  # noqa: B907
        assert "🦘" in body
        if val == "img":
            assert f"/{hex(ord('🦘'))[2:].upper()}.svg" in body

    for bool1, bool2 in (("sure", "true"), ("nope", "false")):
        response = await check_html_page(fetch, f"/?no_3rd_party={bool1}")
        assert response.code == 200
        body = response.body.decode("UTF-8")
        assert "🦘" in body
        response = await check_html_page(fetch, f"/?no_3rd_party={bool2}")
        assert response.code == 200
        assert (
            response.body.decode("UTF-8").replace(
                f"no_3rd_party={bool2}", f"no_3rd_party={bool1}"
            )
            == body
        )


@travel(datetime(2000, 1, 2, 3, 4, 5, 6, hill_valley), tick=False)
async def test_colour_scheme(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the scheme query parameter."""
    assert COLOUR_SCHEMES
    for scheme in COLOUR_SCHEMES:
        response = assert_valid_html_response(await fetch(f"/?scheme={scheme}"))
        assert response.code == 200
        body = response.body.decode("UTF-8")
        if scheme != "system":
            assert f"?scheme={scheme}" in body
        html = HTMLParser(namespaceHTMLElements=False).parse(response.body)
        assert html.find(
            f".[@data-scheme={"light" if scheme == "random" else scheme!r}]"
        ), f"{scheme} should be specified in html"


@travel(datetime.fromtimestamp(1990), tick=False)
async def test_request_handlers2(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the request handlers return 200 codes."""
    response = assert_valid_html_response(await fetch("/"))
    assert response.code == 200
    for cookie_header in response.headers.get_list("Set-Cookie"):
        cookie = parse_cookie(cookie_header)
        assert len(cookie) == 1
        morsel = cookie[tuple(cookie)[0]]
        assert morsel.key == "c"
        assert morsel.value == "s"


test_request_handlers3 = travel("16-4-1")(test_request_handlers2)


async def test_request_handlers4(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the request handlers return 200 codes."""
    response = await fetch("/")
    assert response.code == 200

    assert_valid_html_response(await fetch("/?c=s"))
    response = await check_html_page(
        fetch, "/redirect?from=/&to=https://example.org"
    )
    assert b"https://example.org" in response.body

    assert_valid_response(
        await fetch("/humans.txt"), "text/plain; charset=UTF-8"
    )
    assert_valid_response(
        await fetch("/static/humans.txt"), "text/plain; charset=UTF-8"
    )
    assert_valid_response(
        await fetch("/robots.txt"), "text/plain; charset=UTF-8"
    )
    assert_valid_response(
        await fetch("/static/robots.txt"), "text/plain; charset=UTF-8"
    )
    assert_valid_response(await fetch("/favicon.jxl"), "image/jxl")
    assert_valid_response(await fetch("/static/favicon.jxl"), "image/jxl")
    assert_valid_response(await fetch("/favicon.png"), "image/png")
    assert_valid_response(await fetch("/static/favicon.png"), "image/png")


async def test_request_handlers5(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the request handlers return 200 codes."""
    await check_html_page(fetch, "/betriebszeit", codes={200})
    await check_html_page(fetch, "/discord?ask_before_leaving=1", codes={200})
    await check_html_page(fetch, "/einstellungen", codes={200})
    await check_html_page(fetch, "/endpunkte", codes={200})
    await check_html_page(fetch, "/hangman-loeser", codes={200})
    await check_html_page(fetch, "/host-info", codes={200})
    await check_html_page(fetch, "/ip", codes={200})
    await check_html_page(fetch, "/js-lizenzen", codes={200})
    await check_html_page(fetch, "/kaenguru-comics", codes={200})
    await check_html_page(fetch, "/kontakt", codes={200})
    await check_html_page(fetch, "/services", codes={200})
    await check_html_page(fetch, "/soundboard", codes={200})
    await check_html_page(fetch, "/soundboard/personen", codes={200})
    await check_html_page(fetch, "/soundboard/suche", codes={200})
    await check_html_page(fetch, "/suche", codes={200})
    await check_html_page(fetch, "/version", codes={200})
    await check_html_page(fetch, "/vertauschte-woerter", codes={200})
    await check_html_page(fetch, "/waehrungs-rechner", codes={200})
    await check_html_page(fetch, "/wiki?ask_before_leaving=1", codes={200})
    await check_html_page(fetch, "/wortspiel-helfer", codes={200})


async def test_request_handlers6(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the request handlers return 200 codes."""
    assert_valid_dynload_response(
        await fetch(
            "/betriebszeit",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/discord?ask_before_leaving=sure",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/einstellungen",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/endpunkte",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/hangman-loeser",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/host-info",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/ip", headers={"Accept": "application/vnd.asozial.dynload+json"}
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/js-lizenzen",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/kaenguru-comics",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/kontakt",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/services",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/soundboard",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/soundboard/personen",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/soundboard/suche",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/suche", headers={"Accept": "application/vnd.asozial.dynload+json"}
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/version",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/vertauschte-woerter",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/waehrungs-rechner",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/wiki?ask_before_leaving=sure",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )
    assert_valid_dynload_response(
        await fetch(
            "/wortspiel-helfer",
            headers={"Accept": "application/vnd.asozial.dynload+json"},
        ),
        {200},
    )


async def test_request_handlers7(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the request handlers return 200 codes."""
    response = await fetch("/host-info/uwu")
    assert response.code in {200, 503}
    assert_valid_html_response(response, {response.code})

    assert_valid_rss_response(await fetch("/soundboard/feed"), {200})
    assert_valid_rss_response(await fetch("/soundboard/muk/feed"), {200})
    assert_valid_rss_response(await fetch("/soundboard/qwertzuiop/feed"), {404})

    await check_html_page(fetch, "/soundboard/muk")
    assert_valid_html_response(await fetch("/soundboard/qwertzuiop"), {404})

    assert_valid_response(
        await fetch("/api/backdoor/exec", method="POST", body="42"),
        "application/vnd.uqfoundation.dill",
        {401},
    )

    response = assert_valid_response(
        await fetch("/api/ping"), "text/plain;charset=utf-8"
    )
    assert response.body.decode("UTF-8") == "🏓\n"

    for boolean in (bool(_) for _ in range(2)):
        url = f"/@apm-rum/elastic-apm-rum.umd{'.min' if boolean else ''}.js"
        assert_valid_response(
            await fetch(url, follow_redirects=True),
            "text/javascript; charset=UTF-8",
            needs_to_end_with_line_feed=False,
        )
    assert_valid_response(
        await fetch(url + ".map", follow_redirects=True),  # type: ignore[possibly-undefined]  # noqa: B950
        "text/plain; charset=UTF-8",
        needs_to_end_with_line_feed=False,
    )


async def test_error_code_pages(fetch: FetchCallable) -> None:  # noqa: F811
    """Check if the request handlers return codes."""
    for code in range(200, 599):
        if code not in (204, 304):
            assert_valid_html_response(await fetch(f"/{code}.html"), {code})



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page with wrong quotes."""

from __future__ import annotations

import abc
import asyncio
import contextlib
import logging
import multiprocessing.synchronize
import random
import sys
import time
from collections.abc import (
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    Sequence,
)
from dataclasses import dataclass
from datetime import date
from multiprocessing import Value
from typing import Any, Final, Literal, cast
from urllib.parse import urlencode

import dill  # type: ignore[import-untyped]  # nosec: B403
import elasticapm
import orjson as json
import typed_stream
from redis.asyncio import Redis
from tornado.httpclient import AsyncHTTPClient
from tornado.web import Application, HTTPError
from UltraDict import UltraDict  # type: ignore[import-untyped]

from .. import (
    CA_BUNDLE_PATH,
    DIR as ROOT_DIR,
    EVENT_REDIS,
    EVENT_SHUTDOWN,
    NAME,
    ORJSON_OPTIONS,
    pytest_is_running,
)
from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo, Permission, ratelimit

DIR: Final = ROOT_DIR / "quotes"

LOGGER: Final = logging.getLogger(__name__)

API_URL: Final[str] = "https://zitate.prapsschnalinen.de/api"


# pylint: disable-next=too-few-public-methods
class UltraDictType[K, V](MutableMapping[K, V], abc.ABC):
    """The type of the shared dictionaries."""

    lock: multiprocessing.synchronize.RLock


QUOTES_CACHE: Final[UltraDictType[int, Quote]] = UltraDict(
    buffer_size=1024**2, serializer=dill
)
AUTHORS_CACHE: Final[UltraDictType[int, Author]] = UltraDict(
    buffer_size=1024**2, serializer=dill
)
WRONG_QUOTES_CACHE: Final[UltraDictType[tuple[int, int], WrongQuote]] = (
    UltraDict(buffer_size=1024**2, serializer=dill)
)

MAX_QUOTES_ID = Value("Q", 0)
MAX_AUTHORS_ID = Value("Q", 0)


@dataclass(init=False, slots=True)
class QuotesObjBase(abc.ABC):
    """An object with an id."""

    id: int

    @abc.abstractmethod
    async def fetch_new_data(self) -> QuotesObjBase:
        """Fetch new data from the API."""
        raise NotImplementedError

    # pylint: disable=unused-argument
    def get_id_as_str(self, minify: bool = False) -> str:
        """Get the id of the object as a string."""
        return str(self.id)

    @abc.abstractmethod
    def get_path(self) -> str:
        """Return the path to the Object."""
        raise NotImplementedError


@dataclass(slots=True)
class Author(QuotesObjBase):
    """The author object with a name."""

    name: str
    # tuple(url_to_info, info_str, creation_date)
    info: None | tuple[str, None | str, date]

    def __str__(self) -> str:
        """Return the name of the author."""
        return self.name

    async def fetch_new_data(self) -> Author:
        """Fetch new data from the API."""
        return parse_author(
            await make_api_request(
                f"authors/{self.id}", entity_should_exist=True
            )
        )

    def get_path(self) -> str:
        """Return the path to the author info."""
        return f"/zitate/info/a/{self.id}"

    def to_json(self) -> dict[str, Any]:
        """Get the author as JSON."""
        return {
            "id": self.id,
            "name": str(self),
            "path": self.get_path(),
            "info": (
                {
                    "source": self.info[0],
                    "text": self.info[1],
                    "date": self.info[2].isoformat(),
                }
                if self.info
                else None
            ),
        }


@dataclass(slots=True)
class Quote(QuotesObjBase):
    """The quote object with a quote text and an author."""

    quote: str
    author_id: int

    def __str__(self) -> str:
        """Return the content of the quote."""
        return self.quote.strip()

    @property
    def author(self) -> Author:
        """Get the corresponding author object."""
        return AUTHORS_CACHE[self.author_id]

    async def fetch_new_data(self) -> Quote:
        """Fetch new data from the API."""
        return parse_quote(
            await make_api_request(
                f"quotes/{self.id}", entity_should_exist=True
            ),
            self,
        )

    def get_path(self) -> str:
        """Return the path to the quote info."""
        return f"/zitate/info/z/{self.id}"

    def to_json(self) -> dict[str, Any]:
        """Get the quote as JSON."""
        return {
            "id": self.id,
            "quote": str(self),
            "author": self.author.to_json(),
            "path": self.get_path(),
        }


@dataclass(slots=True)
class WrongQuote(QuotesObjBase):
    """The wrong quote object with a quote, an author and a rating."""

    quote_id: int
    author_id: int
    rating: int

    def __str__(self) -> str:
        r"""
        Return the wrong quote.

        like: '»quote« - author'.
        """
        return f"»{self.quote}« - {self.author}"

    @property
    def author(self) -> Author:
        """Get the corresponding author object."""
        return AUTHORS_CACHE[self.author_id]

    async def fetch_new_data(self) -> WrongQuote:
        """Fetch new data from the API."""
        if self.id == -1:
            api_data = await make_api_request(
                "wrongquotes",
                {
                    "quote": str(self.quote_id),
                    "simulate": "true",
                    "author": str(self.author_id),
                },
                entity_should_exist=True,
            )
            if api_data:
                api_data = api_data[0]
        else:
            api_data = await make_api_request(
                f"wrongquotes/{self.id}", entity_should_exist=True
            )
        if not api_data:
            return self
        return parse_wrong_quote(api_data, self)

    def get_id(self) -> tuple[int, int]:
        """
        Get the id of the quote and the author in a tuple.

        :return tuple(quote_id, author_id)
        """
        return self.quote_id, self.author_id

    def get_id_as_str(self, minify: bool = False) -> str:
        """
        Get the id of the wrong quote as a string.

        Format: quote_id-author_id
        """
        if minify and self.id != -1:
            return str(self.id)
        return f"{self.quote_id}-{self.author_id}"

    def get_path(self) -> str:
        """Return the path to the wrong quote."""
        return f"/zitate/{self.get_id_as_str()}"

    @property
    def quote(self) -> Quote:
        """Get the corresponding quote object."""
        return QUOTES_CACHE[self.quote_id]

    def to_json(self) -> dict[str, Any]:
        """Get the wrong quote as JSON."""
        return {
            "id": self.get_id_as_str(),
            "quote": self.quote.to_json(),
            "author": self.author.to_json(),
            "rating": self.rating,
            "path": self.get_path(),
        }

    async def vote(
        # pylint: disable=unused-argument
        self,
        vote: Literal[-1, 1],
        lazy: bool = False,
    ) -> WrongQuote:
        """Vote for the wrong quote."""
        if self.id == -1:
            raise ValueError("Can't vote for a not existing quote.")
        # if lazy:  # simulate the vote and do the actual voting later
        #     self.rating += vote
        #     asyncio.get_running_loop().call_soon_threadsafe(
        #         self.vote,
        #         vote,
        #     )
        #     return self
        # do the voting
        return parse_wrong_quote(
            await make_api_request(
                f"wrongquotes/{self.id}",
                method="POST",
                body={"vote": str(vote)},
                entity_should_exist=True,
            ),
            self,
        )


def get_wrong_quotes(
    filter_fun: None | Callable[[WrongQuote], bool] = None,
    *,
    sort: bool = False,  # sorted by rating
    filter_real_quotes: bool = True,
    shuffle: bool = False,
) -> Sequence[WrongQuote]:
    """Get cached wrong quotes."""
    if shuffle and sort:
        raise ValueError("Sort and shuffle can't be both true.")
    wqs: list[WrongQuote] = list(WRONG_QUOTES_CACHE.values())
    if filter_fun or filter_real_quotes:
        for i in reversed(range(len(wqs))):
            if (filter_fun and not filter_fun(wqs[i])) or (
                filter_real_quotes
                and wqs[i].quote.author_id == wqs[i].author_id
            ):
                del wqs[i]
    if shuffle:
        random.shuffle(wqs)
    elif sort:
        wqs.sort(key=lambda wq: wq.rating, reverse=True)
    return wqs


def get_quotes(
    filter_fun: None | Callable[[Quote], bool] = None,
    shuffle: bool = False,
) -> list[Quote]:
    """Get cached quotes."""
    quotes: list[Quote] = list(QUOTES_CACHE.values())
    if filter_fun:
        for i in reversed(range(len(quotes))):
            if not filter_fun(quotes[i]):
                del quotes[i]
    if shuffle:
        random.shuffle(quotes)
    return quotes


def get_authors(
    filter_fun: None | Callable[[Author], bool] = None,
    shuffle: bool = False,
) -> list[Author]:
    """Get cached authors."""
    authors: list[Author] = list(AUTHORS_CACHE.values())
    if filter_fun:
        for i in reversed(range(len(authors))):
            if not filter_fun(authors[i]):
                del authors[i]
    if shuffle:
        random.shuffle(authors)
    return authors


async def make_api_request(
    endpoint: str,
    args: Mapping[str, str] | None = None,
    *,
    entity_should_exist: bool,
    method: Literal["GET", "POST"] = "GET",
    body: None | Mapping[str, str] = None,
) -> Any:  # TODO: list[dict[str, Any]] | dict[str, Any]:
    """Make API request and return the result as dict."""
    if pytest_is_running():
        return None
    query = f"?{urlencode(args)}" if args else ""
    url = f"{API_URL}/{endpoint}{query}"
    body_str = urlencode(body) if body else body
    response = await AsyncHTTPClient().fetch(
        url,
        method=method,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        body=body_str,
        raise_error=False,
        ca_certs=CA_BUNDLE_PATH,
    )
    if response.code != 200:
        normed_response_code = (
            400
            if not entity_should_exist and response.code == 500
            else response.code
        )
        LOGGER.log(
            logging.ERROR if normed_response_code >= 500 else logging.WARNING,
            "%s request to %r with body=%r failed with code=%d and reason=%r",
            method,
            url,
            body_str,
            response.code,
            response.reason,
        )
        raise HTTPError(
            normed_response_code if normed_response_code in {400, 404} else 503,
            reason=(
                f"{API_URL}/{endpoint} returned: "
                f"{response.code} {response.reason}"
            ),
        )
    return json.loads(response.body)


def fix_author_name(name: str) -> str:
    """Fix common mistakes in authors."""
    if len(name) > 2 and name.startswith("(") and name.endswith(")"):
        # remove () from author name, that shouldn't be there
        name = name[1:-1]
    return name.strip()


def parse_author(json_data: Mapping[str, Any]) -> Author:
    """Parse an author from JSON data."""
    id_ = int(json_data["id"])
    name = fix_author_name(json_data["author"])

    with AUTHORS_CACHE.lock:
        author = AUTHORS_CACHE.get(id_)
        if author is None:
            # pylint: disable-next=too-many-function-args
            author = Author(id_, name, None)
            MAX_AUTHORS_ID.value = max(MAX_AUTHORS_ID.value, id_)
        elif author.name != name:
            author.name = name
            author.info = None  # reset info

        AUTHORS_CACHE[author.id] = author

    return author


def fix_quote_str(quote_str: str) -> str:
    """Fix common mistakes in quotes."""
    if (
        len(quote_str) > 2
        and quote_str.startswith(('"', "„", "“"))
        and quote_str.endswith(('"', "“", "”"))
    ):
        # remove quotation marks from quote, that shouldn't be there
        quote_str = quote_str[1:-1]

    return quote_str.strip()


def parse_quote(
    json_data: Mapping[str, Any], quote: None | Quote = None
) -> Quote:
    """Parse a quote from JSON data."""
    quote_id = int(json_data["id"])
    author = parse_author(json_data["author"])  # update author
    quote_str = fix_quote_str(json_data["quote"])

    with QUOTES_CACHE.lock:
        if quote is None:  # no quote supplied, try getting it from cache
            quote = QUOTES_CACHE.get(quote_id)
        if quote is None:  # new quote
            # pylint: disable=too-many-function-args
            quote = Quote(quote_id, quote_str, author.id)
            MAX_QUOTES_ID.value = max(MAX_QUOTES_ID.value, quote.id)
        else:  # quote was already saved
            quote.quote = quote_str
            quote.author_id = author.id

        QUOTES_CACHE[quote.id] = quote

    return quote


def parse_wrong_quote(
    json_data: Mapping[str, Any], wrong_quote: None | WrongQuote = None
) -> WrongQuote:
    """Parse a wrong quote and update the cache."""
    quote = parse_quote(json_data["quote"])
    author = parse_author(json_data["author"])

    id_tuple = (quote.id, author.id)
    rating = json_data["rating"]
    wrong_quote_id = int(json_data.get("id") or -1)

    if wrong_quote is None:
        with WRONG_QUOTES_CACHE.lock:
            wrong_quote = WRONG_QUOTES_CACHE.get(id_tuple)
            if wrong_quote is None:
                wrong_quote = (
                    WrongQuote(  # pylint: disable=unexpected-keyword-arg
                        id=wrong_quote_id,
                        quote_id=quote.id,
                        author_id=author.id,
                        rating=rating,
                    )
                )
                WRONG_QUOTES_CACHE[id_tuple] = wrong_quote
                return wrong_quote

    # make sure the wrong quote is the correct one
    if (wrong_quote.quote_id, wrong_quote.author_id) != id_tuple:
        raise HTTPError(reason="ERROR: -41")

    # update the data of the wrong quote
    if wrong_quote.rating != rating:
        wrong_quote.rating = rating
    if wrong_quote.id != wrong_quote_id:
        wrong_quote.id = wrong_quote_id

    WRONG_QUOTES_CACHE[id_tuple] = wrong_quote

    return wrong_quote


async def parse_list_of_quote_data(
    json_list: str | Iterable[Mapping[str, Any]],
    parse_fun: Callable[[Mapping[str, Any]], QuotesObjBase],
) -> tuple[QuotesObjBase, ...]:
    """Parse a list of quote data."""
    if not json_list:
        return ()
    if isinstance(json_list, str):
        json_list = cast(list[dict[str, Any]], json.loads(json_list))
    return_list = []
    for json_data in json_list:
        _ = parse_fun(json_data)
        await asyncio.sleep(0)
        return_list.append(_)
    return tuple(return_list)


async def update_cache_periodically(
    app: Application, worker: int | None
) -> None:
    """Start updating the cache every hour."""
    # pylint: disable=too-complex, too-many-branches
    if "/troet" in typed_stream.Stream(
        cast(Iterable[ModuleInfo], app.settings.get("MODULE_INFOS", ()))
    ).map(lambda m: m.path):
        app.settings["SHOW_SHARING_ON_MASTODON"] = True
    if worker:
        return
    with contextlib.suppress(asyncio.TimeoutError):
        await asyncio.wait_for(EVENT_REDIS.wait(), 5)
    redis: Redis[str] = cast("Redis[str]", app.settings.get("REDIS"))
    prefix: str = app.settings.get("REDIS_PREFIX", NAME).removesuffix("-dev")
    apm: None | elasticapm.Client
    if EVENT_REDIS.is_set():  # pylint: disable=too-many-nested-blocks
        await parse_list_of_quote_data(
            await redis.get(f"{prefix}:cached-quote-data:authors"),  # type: ignore[arg-type]  # noqa: B950
            parse_author,
        )
        await parse_list_of_quote_data(
            await redis.get(f"{prefix}:cached-quote-data:quotes"),  # type: ignore[arg-type]  # noqa: B950
            parse_quote,
        )
        await parse_list_of_quote_data(
            await redis.get(f"{prefix}:cached-quote-data:wrongquotes"),  # type: ignore[arg-type]  # noqa: B950
            parse_wrong_quote,
        )
        if QUOTES_CACHE and AUTHORS_CACHE and WRONG_QUOTES_CACHE:
            last_update = await redis.get(
                f"{prefix}:cached-quote-data:last-update"
            )
            if last_update:
                last_update_int = int(last_update)
                since_last_update = int(time.time()) - last_update_int
                if 0 <= since_last_update < 60 * 60:
                    # wait until the last update is at least one hour old
                    update_cache_in = 60 * 60 - since_last_update
                    if not sys.flags.dev_mode and update_cache_in > 60:
                        # if in production mode update wrong quotes just to be sure
                        try:
                            await update_cache(
                                app, update_quotes=False, update_authors=False
                            )
                        except Exception:  # pylint: disable=broad-except
                            LOGGER.exception("Updating quotes cache failed")
                            apm = app.settings.get("ELASTIC_APM", {}).get(
                                "CLIENT"
                            )
                            if apm:
                                apm.capture_exception()
                        else:
                            LOGGER.info("Updated quotes cache successfully")
                    LOGGER.info(
                        "Next update of quotes cache in %d seconds",
                        update_cache_in,
                    )
                    await asyncio.sleep(update_cache_in)

    # update the cache every hour
    failed = 0
    while not EVENT_SHUTDOWN.is_set():  # pylint: disable=while-used
        try:
            await update_cache(app)
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("Updating quotes cache failed")
            if apm := app.settings.get("ELASTIC_APM", {}).get("CLIENT"):
                apm.capture_exception()
            failed += 1
            await asyncio.sleep(pow(min(failed * 2, 60), 2))  # 4,16,...,60*60
        else:
            LOGGER.info("Updated quotes cache successfully")
            failed = 0
            await asyncio.sleep(60 * 60)


async def update_cache(
    app: Application,
    update_wrong_quotes: bool = True,
    update_quotes: bool = True,
    update_authors: bool = True,
) -> None:
    """Fill the cache with all data from the API."""
    LOGGER.info("Updating quotes cache")
    redis: Redis[str] = cast("Redis[str]", app.settings.get("REDIS"))
    prefix: str = app.settings.get("REDIS_PREFIX", NAME).removesuffix("-dev")
    redis_available = EVENT_REDIS.is_set()

    if update_wrong_quotes:
        await parse_list_of_quote_data(
            wq_data := await make_api_request(
                "wrongquotes", entity_should_exist=True
            ),
            parse_wrong_quote,
        )
        if wq_data and redis_available:
            await redis.setex(
                f"{prefix}:cached-quote-data:wrongquotes",
                60 * 60 * 24 * 30,
                json.dumps(wq_data, option=ORJSON_OPTIONS),
            )

    if update_quotes:
        await parse_list_of_quote_data(
            quotes_data := await make_api_request(
                "quotes", entity_should_exist=True
            ),
            parse_quote,
        )
        if quotes_data and redis_available:
            await redis.setex(
                f"{prefix}:cached-quote-data:quotes",
                60 * 60 * 24 * 30,
                json.dumps(quotes_data, option=ORJSON_OPTIONS),
            )

    if update_authors:
        await parse_list_of_quote_data(
            authors_data := await make_api_request(
                "authors", entity_should_exist=True
            ),
            parse_author,
        )
        if authors_data and redis_available:
            await redis.setex(
                f"{prefix}:cached-quote-data:authors",
                60 * 60 * 24 * 30,
                json.dumps(authors_data, option=ORJSON_OPTIONS),
            )

    if (
        redis_available
        and update_wrong_quotes
        and update_quotes
        and update_authors
    ):
        await redis.setex(
            f"{prefix}:cached-quote-data:last-update",
            60 * 60 * 24 * 30,
            int(time.time()),
        )


async def get_author_by_id(author_id: int) -> Author:
    """Get an author by its id."""
    author = AUTHORS_CACHE.get(author_id)
    if author is not None:
        return author
    return parse_author(
        await make_api_request(
            f"authors/{author_id}", entity_should_exist=False
        )
    )


async def get_quote_by_id(quote_id: int) -> Quote:
    """Get a quote by its id."""
    quote = QUOTES_CACHE.get(quote_id)
    if quote is not None:
        return quote
    return parse_quote(
        await make_api_request(f"quotes/{quote_id}", entity_should_exist=False)
    )


async def get_wrong_quote(
    quote_id: int, author_id: int, use_cache: bool = True
) -> WrongQuote:
    """Get a wrong quote with a quote id and an author id."""
    wrong_quote = WRONG_QUOTES_CACHE.get((quote_id, author_id))
    if wrong_quote:
        if use_cache:
            return wrong_quote
        # do not use cache, so update the wrong quote data
        return await wrong_quote.fetch_new_data()
    # wrong quote not in cache
    if use_cache and quote_id in QUOTES_CACHE and author_id in AUTHORS_CACHE:
        # we don't need to request anything, as the wrong_quote probably has
        # no ratings just use the cached quote and author
        # pylint: disable-next=too-many-function-args
        return WrongQuote(-1, quote_id, author_id, 0)
    # request the wrong quote from the API
    result = await make_api_request(
        "wrongquotes",
        {
            "quote": str(quote_id),
            "simulate": "true",
            "author": str(author_id),
        },
        entity_should_exist=False,
    )
    if result:
        return parse_wrong_quote(result[0])

    raise HTTPError(404, reason="Falsches Zitat nicht gefunden")


async def get_rating_by_id(quote_id: int, author_id: int) -> int:
    """Get the rating of a wrong quote."""
    return (await get_wrong_quote(quote_id, author_id)).rating


def get_random_quote_id() -> int:
    """Get random quote id."""
    return random.randint(1, MAX_QUOTES_ID.value)  # nosec: B311


def get_random_author_id() -> int:
    """Get random author id."""
    return random.randint(1, MAX_AUTHORS_ID.value)  # nosec: B311


def get_random_id() -> tuple[int, int]:
    """Get random wrong quote id."""
    return (
        get_random_quote_id(),
        get_random_author_id(),
    )


async def create_wq_and_vote(
    vote: Literal[-1, 1],
    quote_id: int,
    author_id: int,
    contributed_by: str,
    fast: bool = False,
) -> WrongQuote:
    """
    Vote for the wrong_quote with the API.

    If the wrong_quote doesn't exist yet, create it.
    """
    wrong_quote = WRONG_QUOTES_CACHE.get((quote_id, author_id))
    if wrong_quote and wrong_quote.id != -1:
        return await wrong_quote.vote(vote, fast)
    # we don't know the wrong_quote_id, so we have to create the wrong_quote
    wrong_quote = parse_wrong_quote(
        await make_api_request(
            "wrongquotes",
            method="POST",
            body={
                "quote": str(quote_id),
                "author": str(author_id),
                "contributed_by": contributed_by,
            },
            entity_should_exist=False,
        )
    )
    return await wrong_quote.vote(vote, lazy=True)


class QuoteReadyCheckHandler(HTMLRequestHandler):
    """Class that checks if quotes have been loaded."""

    async def check_ready(self) -> None:
        """Fail if quotes aren't ready yet."""
        if not WRONG_QUOTES_CACHE:
            # should work in a few seconds, the quotes just haven't loaded yet
            self.set_header("Retry-After", "5")
            raise HTTPError(503, reason="Service available in a few seconds")

    async def prepare(self) -> None:  # noqa: D102
        await super().prepare()
        if self.request.method != "OPTIONS":
            await self.check_ready()

        if (  # pylint: disable=too-many-boolean-expressions
            self.settings.get("RATELIMITS")
            and self.request.method not in {"HEAD", "OPTIONS"}
            and not self.is_authorized(Permission.RATELIMITS)
            and not self.crawler
            and (
                self.request.path.endswith(".xlsx")
                or self.content_type == "application/vnd.ms-excel"
            )
        ):
            if self.settings.get("UNDER_ATTACK") or not EVENT_REDIS.is_set():
                raise HTTPError(503)

            ratelimited, headers = await ratelimit(
                self.redis,
                self.redis_prefix,
                str(self.request.remote_ip),
                bucket="quotes:image:xlsx",
                max_burst=4,
                count_per_period=1,
                period=60,
                tokens=1 if self.request.method != "HEAD" else 0,
            )

            for header, value in headers.items():
                self.set_header(header, value)

            if ratelimited:
                if self.now.date() == date(self.now.year, 4, 20):
                    self.set_status(420)
                    self.write_error(420)
                else:
                    self.set_status(429)
                    self.write_error(429)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
The redirect page of the website.

This page is used to redirect users to third party websites.
This page will ask users if they want to leave this website.
"""

from __future__ import annotations

from urllib.parse import urlsplit

from tornado.web import Application

from .. import pytest_is_running
from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/redirect", RedirectPage),),
        name="Weiterleitungsseite",
        description=(
            "Seite, die User davon abhält versehentlich eine fremde "
            "Webseite zu besuchen"
        ),
        path="/redirect",
        short_name="Weiterleitung",
        hidden=True,
        required_background_tasks=(_confirm_loading,),
    )


async def _confirm_loading(app: Application, worker: int | None) -> None:
    """Save in settings that the redirect module has been loaded."""
    app.settings["REDIRECT_MODULE_LOADED"] = worker or True  # noqa: SIM222


class RedirectPage(HTMLRequestHandler):
    """The redirect page that redirects you to another page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the redirect page."""
        # pylint: disable=unused-argument

        redirect_url = self.get_argument("to", None) or "/"
        referrer = self.request.headers.get_list("Referer")
        from_url = referrer[0] if referrer else None

        if redirect_url.startswith("/"):
            # it is a local URL, so just redirect
            # use fix_url to maybe add no_3rd_party
            return self.redirect(self.fix_url(redirect_url))

        if (
            (domain := urlsplit(redirect_url).hostname)
            and domain.endswith(".asozial.org")
            and not pytest_is_running()
        ):
            return self.redirect(redirect_url)

        await self.render(
            "pages/redirect.html",
            redirect_url=redirect_url,
            send_referrer=False,
            from_url=from_url,
            discord=False,
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the token module."""

from __future__ import annotations

from datetime import datetime

import pytest

from an_website.utils.token import (  # pylint: disable=import-private-name
    InvalidTokenError,
    _create_token_body_v0,
    _parse_token_v0,
    bytes_to_int,
    create_token,
    int_to_bytes,
    parse_token,
)
from an_website.utils.utils import Permission


def test_token_v0() -> None:
    """Test the token creation."""
    result = create_token(Permission(1), secret=b"xyzzy", duration=9999)
    assert result.permissions == Permission(1)
    token = _create_token_body_v0(
        Permission(20),
        b"xyzzy",
        100,
        datetime.fromtimestamp(0),
        b"\x00\x00\x00\x00\x00*",  # 42
    )
    assert (
        token == "ABQAAAAAACoAAAAAZAAAAAAAVuGjEJjux2P8zah"  # nosec: B105
        "9SjQGkxCWn9Iaszr/qJJHJSMWoRPoVjQKa6/jKFWPdBehSL0K"
    )

    parsed = _parse_token_v0(token, b"xyzzy", verify_time=False)
    assert parsed.salt == b"\x00\x00\x00\x00\x00*"
    assert parsed.permissions == Permission(20)
    assert parsed.valid_until == datetime.fromtimestamp(100)

    with pytest.raises(InvalidTokenError):
        _parse_token_v0(token, b"xyzzy")  # expired

    with pytest.raises(InvalidTokenError):
        parse_token("0" + token, secret=b"xyzzy")

    with pytest.raises(InvalidTokenError):
        _parse_token_v0(token, b"hunter2", verify_time=False)


def test_int_to_bytes() -> None:
    """Test the int to bytes conversion."""
    assert int_to_bytes(0, 2) == b"\0" * 2
    assert int_to_bytes(0x20, 2) == b"\0 "
    assert int_to_bytes(0x2000, 2) == b" \0"


def test_bytes_to_int() -> None:
    """Test the bytes to int conversion."""
    assert bytes_to_int(b"\0") == 0  # pylint: disable=compare-to-zero
    assert bytes_to_int(b" ") == 0x20
    assert bytes_to_int(b"  ") == 0x2020
    assert bytes_to_int(b" \0") == 0x2000



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The API for updating the website."""

from __future__ import annotations

import asyncio
import logging
import os
import sys
from asyncio import Future
from queue import SimpleQueue
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import IO, TYPE_CHECKING, Any, ClassVar, Final
from urllib.parse import unquote

from tornado.web import stream_request_body

from .. import EVENT_SHUTDOWN, NAME
from ..utils.decorators import requires
from ..utils.request_handler import APIRequestHandler
from ..utils.utils import ModuleInfo, Permission

if TYPE_CHECKING:
    from tempfile import _TemporaryFileWrapper

LOGGER: Final = logging.getLogger(__name__)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/api/update/(.*)", UpdateAPI),),
        name="Update-API",
        description=f"API zum Aktualisieren von {NAME.removesuffix('-dev')}",
        hidden=True,
    )


def write_from_queue(file: IO[bytes], queue: SimpleQueue[None | bytes]) -> None:
    """Read from a queue and write to a file."""
    while (chunk := queue.get()) is not None:  # pylint: disable=while-used
        file.write(chunk)
    file.flush()


@stream_request_body
class UpdateAPI(APIRequestHandler):  # pragma: no cover
    """The request handler for the update API."""

    ALLOWED_METHODS: ClassVar[tuple[str, ...]] = ("PUT",)
    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = ("text/plain",)

    dir: TemporaryDirectory[str]
    file: _TemporaryFileWrapper[bytes]
    queue: SimpleQueue[None | bytes]
    future: Future[Any]

    def data_received(self, chunk: bytes) -> None:  # noqa: D102
        self.queue.put(chunk)

    def on_finish(self) -> None:  # noqa: D102
        if hasattr(self, "queue"):
            self.queue.put(None)

    async def pip_install(self, *args: str) -> int:
        """Install something and write the output."""
        process = await asyncio.create_subprocess_exec(
            sys.executable,
            "-m",
            "pip",
            "install",
            "--require-virtualenv",
            *args,
            stdin=asyncio.subprocess.DEVNULL,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        # pylint: disable=while-used
        while not process.stdout.at_eof():  # type: ignore[union-attr]
            char = await process.stdout.read(1)  # type: ignore[union-attr]
            self.write(char)
            if char == b"\n":
                self.flush()  # type: ignore[unused-awaitable]
        await process.wait()
        assert process.returncode is not None
        return process.returncode

    async def prepare(self) -> None:  # noqa: D102
        await super().prepare()
        loop = asyncio.get_running_loop()
        self.dir = TemporaryDirectory()
        self.file = NamedTemporaryFile(dir=self.dir.name, delete=False)
        self.queue = SimpleQueue()
        self.future = loop.run_in_executor(
            None, write_from_queue, self.file, self.queue
        )

    @requires(Permission.UPDATE)
    async def put(self, filename: str) -> None:
        """Handle PUT requests to the update API."""
        self.queue.put(None)
        await self.future
        self.file.close()

        filepath = os.path.join(self.dir.name, unquote(filename))
        os.rename(self.file.name, filepath)

        self.set_status(202)
        self.set_header("X-Accel-Buffering", "no")

        await self.pip_install("--upgrade", "pip")

        returncode = await self.pip_install(filepath)

        await self.finish()

        if returncode:
            LOGGER.error("Failed to install %s", filename)
        elif self.get_bool_argument("shutdown", True):
            EVENT_SHUTDOWN.set()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the TraversableStaticFileHandler."""

from __future__ import annotations

import gzip

import zstandard

from an_website import DIR as ROOT_DIR
from an_website.utils.fix_static_path_impl import recurse_directory

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_response,
    fetch,
)

STATIC_DIR = ROOT_DIR / "static"
CACHE_CONTROL = f"public, immutable, max-age={86400 * 365 * 10}"


async def test_well_known(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the /.well-known handler."""
    assert_valid_response(
        await fetch("/.well-known/test-file"),
        content_type="text/plain;charset=utf-8",
        codes={404},
        headers={"Access-Control-Allow-Origin": "*"},
    )


async def test_openmoji(fetch: FetchCallable) -> None:  # noqa: F811
    """Test requesting an OpenMoji svg."""
    response = assert_valid_response(
        await fetch("/static/openmoji/svg/1FAE9.svg?v=16.0.0"),
        content_type="image/svg+xml",
        codes={200},
    )
    assert response.headers["Cache-Control"] == CACHE_CONTROL
    size = len(response.body)

    for encoding in ("gzip", "zstd"):
        compressed_response = await fetch(
            "/static/openmoji/svg/1FAE9.svg?v=16.0.0",
            headers={"Accept-Encoding": encoding},
        )
        assert size > (size_compre := len(compressed_response.body))
        assert compressed_response.headers["Content-Encoding"] == encoding
        assert int(compressed_response.headers["Content-Length"]) == size_compre
        if encoding == "gzip":
            assert response.body == gzip.decompress(compressed_response.body)
        if encoding == "zstd":
            assert response.body == zstandard.decompress(
                compressed_response.body
            )

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1FAE9.svg?v=16.0.0",
            headers={"Range": "bytes=17-41"},
        ),
        content_type="image/svg+xml",
        codes={206},
        needs_to_end_with_line_feed=False,
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert part_response.headers["Content-Range"] == f"bytes 17-41/{size}"
    assert int(part_response.headers["Content-Length"]) == 25

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1FAE9.svg?v=16.0.0",
            headers={"Range": "bytes=0-3"},
        ),
        content_type="image/svg+xml",
        codes={206},
        needs_to_end_with_line_feed=False,
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert part_response.headers["Content-Range"] == f"bytes 0-3/{size}"
    assert int(part_response.headers["Content-Length"]) == 4
    assert part_response.body == b"<svg"

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1FAE9.svg?v=16.0.0",
            headers={"Range": f"bytes={size - 7}-{size}"},
        ),
        content_type="image/svg+xml",
        codes={206},
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert (
        part_response.headers["Content-Range"]
        == f"bytes {size - 7}-{size - 1}/{size}"
    )
    assert int(part_response.headers["Content-Length"]) == 7
    assert part_response.body == b"</svg>\n"

    part_response = assert_valid_response(
        await fetch(
            "/static/openmoji/svg/1FAE9.svg?v=16.0.0",
            headers={"Range": "bytes=-7"},
        ),
        content_type="image/svg+xml",
        codes={206},
    )
    assert part_response.headers["Cache-Control"] == CACHE_CONTROL
    assert (
        part_response.headers["Content-Range"]
        == f"bytes {size - 7}-{size - 1}/{size}"
    )
    assert int(part_response.headers["Content-Length"]) == 7
    assert part_response.body == b"</svg>\n"


async def test_static_file_compression(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test fetching static files."""
    file_count, gzip_count, zstd_count = 0, 0, 0

    for file in recurse_directory(
        STATIC_DIR,
        filter=lambda p: p.is_file() and not p.name.endswith((".zst", ".gz")),
    ):
        gzip_file = (STATIC_DIR / f"{file}.gz").is_file()
        zstd_file = (STATIC_DIR / f"{file}.zst").is_file()

        file_count += 1
        gzip_count += gzip_file
        zstd_count += zstd_file

        uncompressed_body = b""

        for encoding in ("identity", "gzip", "zstd"):
            headers = {"Accept-Encoding": encoding}
            response = await fetch(f"/static/{file}", headers=headers)
            body = response.body
            assert response.code == 200
            assert response.headers["Content-Length"] == str(len(body))

            head_response = await fetch(
                f"/static/{file}", headers=headers, method="HEAD"
            )
            assert head_response.code == 200
            assert head_response.headers["Content-Length"] == str(len(body))

            headers = dict(response.headers)
            head_headers = dict(head_response.headers)
            del head_headers["Date"], headers["Date"]
            assert head_headers == headers, f"{head_headers} != {headers}"

            if encoding == "identity":
                assert "Content-Encoding" not in response.headers
                uncompressed_body = body
                continue
            if "Content-Encoding" not in response.headers:
                if encoding == "gzip":
                    assert not gzip_file
                elif encoding == "zstd":
                    assert not zstd_file
                else:
                    raise AssertionError(f"Unknown encoding {encoding}")
                continue

            assert response.headers.get("Content-Encoding") == encoding

            body = response.body
            assert len(body) < len(uncompressed_body)

            decompressed_body = b""
            if encoding == "gzip":
                assert gzip_file
                decompressed_body = gzip.decompress(body)
            elif encoding == "zstd":
                assert zstd_file
                decompressed_body = zstandard.decompress(body)
            else:
                raise AssertionError(f"Unknown encoding {encoding}")

            assert decompressed_body
            assert uncompressed_body == decompressed_body

    assert file_count >= gzip_count > 0
    assert file_count >= zstd_count > 0


async def test_invalid_paths(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test various different invalid paths."""
    response = await fetch("/static/humans.txt/")
    assert response.code == 307
    assert response.headers["location"].endswith("/static/humans.txt")

    response = await fetch("/static//humans.txt")
    assert response.code == 404

    response = await fetch("/static/INVALID_FILE")
    assert response.code == 404

    response = await fetch("/static/HUMANS.TXT")
    assert response.code == 307
    assert response.headers["location"].endswith("/static/humans.txt")


async def test_invalid_range(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test invalid range requests."""
    for encoding in ("identity", "gzip", "zstd"):
        headers = {"Accept-Encoding": encoding}

        response = await fetch("/static/robots.txt", headers=headers)
        size = len(response.body)

        response = await fetch(
            "/static/robots.txt", headers={**headers, "Range": f"bytes={size}"}
        )
        assert response.code == 416
        assert response.headers["Content-Type"] == "text/plain"
        assert response.headers["Content-Range"] == f"bytes */{size}"

        response = await fetch(
            "/static/robots.txt",
            headers={**headers, "Range": f"bytes={size * 2 + 100}"},
        )
        assert response.code == 416
        assert response.headers["Content-Type"] == "text/plain"
        assert response.headers["Content-Range"] == f"bytes */{size}"

        response = await fetch(
            "/static/robots.txt", headers={**headers, "Range": "bytes=-0"}
        )
        assert response.code == 416
        assert response.headers["Content-Type"] == "text/plain"
        assert response.headers["Content-Range"] == f"bytes */{size}"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page that helps solving hangman puzzles."""

from __future__ import annotations

from dataclasses import dataclass

from hangman_solver import (
    HangmanResult,
    Language,
    UnknownLanguageError,
    read_words_with_length,
    solve,
    solve_crossword,
)
from tornado.web import HTTPError

from ..utils.base_request_handler import BaseRequestHandler
from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    languages = "|".join([lang.value.lower() for lang in Language.values()])
    return ModuleInfo(
        handlers=(
            (r"/hangman-loeser", HangmanSolver),
            (r"/api/hangman-loeser", HangmanSolverAPI),
            (
                rf"/hangman-loeser/worte/({languages})/([1-9]\d*).txt",
                HangmanSolverWords,
            ),
        ),
        name="Hangman-Löser",
        description="Eine Webseite, die Lösungen für Galgenmännchen findet",
        path="/hangman-loeser",
        keywords=("Galgenmännchen", "Hangman", "Löser", "Solver", "Worte"),
        aliases=(
            "/hangman-l%C3%B6ser",
            "/hangman-löser",
            "/hangman-solver",
        ),
    )


@dataclass(slots=True)
class HangmanArguments:
    """Arguments for the hangman solver."""

    max_words: int = 20
    crossword_mode: bool = False
    lang: str = "de"
    input: str = ""
    invalid: str = ""

    def get_max_words(self) -> int:
        """Return the maximum number of words."""
        return max(0, min(100, self.max_words))


def solve_hangman(data: HangmanArguments) -> HangmanResult:
    """Generate a hangman object based on the input and return it."""
    try:
        lang = Language.parse_string(data.lang)
    except UnknownLanguageError as err:
        raise HTTPError(
            400, reason=f"{data.lang!r} is an invalid language"
        ) from err

    return (solve_crossword if data.crossword_mode else solve)(
        data.input, data.invalid, lang, data.get_max_words()
    )


class HangmanSolver(HTMLRequestHandler):
    """Request handler for the hangman solver page."""

    RATELIMIT_GET_LIMIT = 10

    @parse_args(type_=HangmanArguments, name="data")
    async def get(self, *, data: HangmanArguments, head: bool = False) -> None:
        """Handle GET requests to the hangman solver page."""
        if head:
            return

        await self.render(
            "pages/hangman_solver.html",
            hangman_result=solve_hangman(data),
            data=data,
        )


class HangmanSolverAPI(APIRequestHandler, HangmanSolver):
    """Request handler for the hangman solver API."""

    RATELIMIT_GET_LIMIT = 10

    @parse_args(type_=HangmanArguments, name="data")
    async def get(self, *, data: HangmanArguments, head: bool = False) -> None:
        """Handle GET requests to the hangman solver API."""
        if head:
            return
        hangman_result = solve_hangman(data)
        await self.finish(
            {
                "input": hangman_result.input,
                "invalid": "".join(hangman_result.invalid),
                "words": hangman_result.words,
                "word_count": hangman_result.matching_words_count,
                "letters": dict(hangman_result.letter_frequency),
                "crossword_mode": data.crossword_mode,
                "max_words": data.get_max_words(),
                "lang": hangman_result.language.value,
            }
        )


class HangmanSolverWords(BaseRequestHandler):
    """Request handler for the hangman word lists."""

    RATELIMIT_GET_LIMIT = 15
    POSSIBLE_CONTENT_TYPES = ("text/plain",)

    async def get(
        self, language: str, length: str, *, head: bool = False
    ) -> None:
        """Handle GET requests to the hangman solver page."""
        # pylint: disable=unused-argument
        try:
            lang = Language.parse_string(language)
        except UnknownLanguageError as err:
            raise HTTPError(404) from err

        try:
            word_length = int(length)
        except ValueError as err:
            raise HTTPError(404) from err

        for word in read_words_with_length(lang, word_length):
            self.write(word)
            self.write(b"\n")

        await self.finish()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page that allows users to contact the website operator."""

from __future__ import annotations

import asyncio
import logging
import smtplib
import ssl
import sys
import time
from collections.abc import Iterable, Mapping
from configparser import ConfigParser
from datetime import datetime, timezone
from email import utils as email_utils
from email.message import Message
from math import pi
from typing import Any, Final, cast
from urllib.parse import quote, urlencode

import typed_stream
from tornado.web import Application, HTTPError, MissingArgumentError

from .. import NAME
from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo

LOGGER: Final = logging.getLogger(__name__)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/kontakt", ContactPage),),
        name="Kontakt-Formular",
        description="Nehme mit dem Betreiber der Webseite Kontakt auf.",
        path="/kontakt",
        keywords=("Kontakt", "Formular"),
        aliases=("/contact",),
        hidden=True,
    )


def apply_contact_stuff_to_app(app: Application, config: ConfigParser) -> None:
    """Apply contact stuff to the app."""
    if "/kontakt" not in typed_stream.Stream(
        cast(Iterable[ModuleInfo], app.settings.get("MODULE_INFOS", ()))
    ).map(lambda m: m.path):
        return

    contact_address = config.get(
        "CONTACT",
        "CONTACT_ADDRESS",
        fallback=(
            f"{NAME.removesuffix('-dev')}@restmail.net"
            if app.settings.get("debug")
            else ""
        ),
    )
    sender_address = config.get(
        "CONTACT",
        "SENDER_ADDRESS",
        fallback=(
            "Marcell D'Avis <davis@1und1.de>"
            if contact_address.endswith("@restmail.net")
            else None
        ),
    )

    if not sender_address:
        # the contact form in other mode if no sender address is set
        app.settings.update(
            CONTACT_ADDRESS=contact_address,
            CONTACT_USE_FORM=1,
        )
        return
    app.settings["CONTACT_ADDRESS"] = None

    if not (contact_address and sender_address):
        return

    app.settings.update(
        CONTACT_USE_FORM=2,
        CONTACT_RECIPIENTS={
            address.strip() for address in contact_address.split(",")
        },
        CONTACT_SMTP_SERVER=config.get(
            "CONTACT",
            "SMTP_SERVER",
            fallback=(
                "restmail.net"
                if contact_address.endswith("@restmail.net")
                else "localhost"
            ),
        ),
        CONTACT_SMTP_PORT=config.getint(
            "CONTACT",
            "SMTP_PORT",
            fallback=25 if contact_address.endswith("@restmail.net") else 587,
        ),
        CONTACT_SMTP_STARTTLS=config.getboolean(
            "CONTACT",
            "SMTP_STARTTLS",
            fallback=None,
        ),
        CONTACT_SENDER_ADDRESS=sender_address,
        CONTACT_SENDER_USERNAME=config.get(
            "CONTACT", "SENDER_USERNAME", fallback=None
        ),
        CONTACT_SENDER_PASSWORD=config.get(
            "CONTACT", "SENDER_PASSWORD", fallback=None
        ),
    )


def send_message(  # pylint: disable=too-many-arguments
    message: Message,
    from_address: str,
    recipients: Iterable[str],
    server: str = "localhost",
    sender: None | str = None,
    username: None | str = None,
    password: None | str = None,
    starttls: None | bool = None,
    port: int = 587,
    *,
    date: None | datetime = None,
) -> dict[str, tuple[int, bytes]]:
    """Send an email."""
    recipients = list(recipients)
    for spam, eggs in enumerate(recipients):
        if eggs.startswith("@"):
            recipients[spam] = "contact" + eggs

    message["Date"] = email_utils.format_datetime(
        date or datetime.now(tz=timezone.utc)
    )

    if sender:
        message["Sender"] = sender
    message["From"] = from_address
    message["To"] = ", ".join(recipients)

    with smtplib.SMTP(server, port) as smtp:
        smtp.set_debuglevel(sys.flags.dev_mode * 2)
        smtp.ehlo_or_helo_if_needed()
        if starttls is None:
            starttls = smtp.has_extn("starttls")
        if starttls:
            smtp.starttls(context=ssl.create_default_context())
        if username and password:
            smtp.login(username, password)
        return smtp.send_message(message)


def add_geoip_info_to_message(
    message: Message,
    geoip_info: Mapping[str, Any],
    header_prefix: str = "X-GeoIP",
) -> None:
    """Add GeoIP information to the message."""
    for spam, eggs in geoip_info.items():
        header = f"{header_prefix}-{spam.replace('_', '-')}"
        if isinstance(eggs, dict):
            add_geoip_info_to_message(message, eggs, header)
        else:
            message[header] = str(eggs)


class ContactPage(HTMLRequestHandler):
    """The request handler for the contact page."""

    RATELIMIT_POST_LIMIT = 5
    RATELIMIT_POST_COUNT_PER_PERIOD = 1
    RATELIMIT_POST_PERIOD = 120

    ACCESS_LOG: dict[str, float] = {}

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the contact page."""
        if not self.settings.get("CONTACT_USE_FORM"):
            raise HTTPError(503)
        if head:
            return

        current_time = time.monotonic()

        # clean up old items from the access log
        for key, value in tuple(self.ACCESS_LOG.items()):
            if current_time - value > 10:
                del self.ACCESS_LOG[key]

        self.ACCESS_LOG[str(self.request.remote_ip)] = current_time

        await self.render(
            "pages/contact.html",
            subject=self.get_argument("subject", ""),
            message=self.get_argument("message", ""),
        )

    async def post(self) -> None:
        """Handle POST requests to the contact page."""
        if not self.settings.get("CONTACT_USE_FORM"):
            raise HTTPError(503)

        if atime := self.ACCESS_LOG.get(str(self.request.remote_ip)):
            del self.ACCESS_LOG[str(self.request.remote_ip)]
            if time.monotonic() - atime < pi:
                LOGGER.info("Rejected message because of timing")
                await self.render(
                    "pages/empty.html",
                    text="Nicht gesendet, da du zu schnell warst.",
                )
                return

        text = self.get_argument("nachricht")
        if not text:
            raise MissingArgumentError("nachricht")  # raise on empty message

        name = self.get_argument("name", "")
        address_arg = self.get_argument("addresse", "")
        address = (
            f"{name} <{address_arg or 'anonymous@foo.bar'}>"
            if name
            else address_arg or "anonymous@foo.bar"
        )

        message = Message()
        host = self.request.host_name
        name = name or address_arg or "Jemand"
        message["Subject"] = str(
            self.get_argument("subjekt", "")
            or f"{name} will etwas über {host} schreiben."
        )
        message.set_payload(text, "UTF-8")

        if honeypot := self.get_argument("message", ""):  # 🍯
            LOGGER.info(
                "rejected message: %s",
                {
                    "Subject": message["Subject"],
                    "message": message,
                    "address": address,
                    "geoip": await self.geoip(),
                    "🍯": honeypot,
                },
            )
            await self.render("pages/empty.html", text="Erfolgreich gesendet.")
            return

        if self.settings.get("CONTACT_USE_FORM") == 1:
            query = urlencode(
                {
                    "subject": message["Subject"],
                    "body": f"{text}\n\nVon: {address}",
                },
                quote_via=quote,
            )
            self.redirect(
                f"mailto:{self.settings.get('CONTACT_ADDRESS')}?{query}",
                status=303,
            )
            return

        geoip = await self.geoip()
        if geoip:
            add_geoip_info_to_message(message, geoip)

        await asyncio.to_thread(
            send_message,
            message=message,
            from_address=address,
            server=self.settings.get("CONTACT_SMTP_SERVER", "localhost"),
            sender=self.settings.get("CONTACT_SENDER_ADDRESS"),
            recipients=self.settings.get("CONTACT_RECIPIENTS", ""),
            username=self.settings.get("CONTACT_SENDER_USERNAME"),
            password=self.settings.get("CONTACT_SENDER_PASSWORD"),
            starttls=self.settings.get("CONTACT_SMTP_STARTTLS"),
            port=self.settings.get("CONTACT_SMTP_PORT", 587),
            date=await self.get_time(),
        )

        await self.render("pages/empty.html", text="Erfolgreich gesendet.")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the backdoor."""

from __future__ import annotations

import pickle  # nosec: B403
from collections.abc import Callable
from functools import partial
from types import EllipsisType
from typing import Any, Literal, TypedDict, cast

import dill  # type: ignore[import-untyped]  # nosec: B403
import jsonpickle  # type: ignore[import-untyped]

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_response,
    fetch,
)

Result = tuple[str, Any]  # tuple[str, bytes] before unpickling


class Response(TypedDict):  # noqa: D101
    # pylint: disable=missing-class-docstring
    success: bool | EllipsisType
    output: None | str
    result: None | Result | SystemExit


ErrorTuple = tuple[int, str]


async def _request_and_parse(
    # pylint: disable=redefined-outer-name, too-many-arguments
    fetch: FetchCallable,  # noqa: F811
    /,
    code: str,
    *,
    accept: str,
    loads: Callable[[bytes], Any],
    content_type: str | None = None,
    session: None | str = None,
    auth_key: None | str = None,
    mode: Literal["exec", "eval"] = "eval",
    features: None | str = None,
) -> Response | ErrorTuple | str | None:
    """Make a request to the backdoor and parse the response."""
    auth_key = "123qweQWE!@#000000000" if auth_key is None else auth_key
    headers = {"Authorization": auth_key, "Accept": accept}

    if session:
        headers["X-Backdoor-Session"] = session

    if features:
        headers["X-Future-Feature"] = features

    http_response = await fetch(
        f"/api/backdoor/{mode}",
        method="POST",
        headers=headers,
        body=code,
    )

    assert_valid_response(http_response, content_type or accept)

    response: Response | ErrorTuple | str | None = loads(http_response.body)

    assert isinstance(response, dict | tuple | str | None)

    if not isinstance(response, dict):
        if isinstance(response, tuple):
            assert len(response) == 2
            assert isinstance(response[0], int)
            assert isinstance(response[1], str)
        return response

    assert len(response) == 3
    assert isinstance(response["success"], bool | EllipsisType)
    assert isinstance(response["output"], None | str)
    assert isinstance(response["result"], None | tuple | SystemExit)
    if isinstance(response["result"], tuple):
        assert len(response["result"]) == 2
        assert isinstance(response["result"][0], str)
        assert isinstance(response["result"][1], bytes | None)
        if isinstance(response["result"][1], bytes):
            result = list(response["result"])
            result[1] = loads(result[1])
            response["result"] = cast(Result, tuple(result))

    return response


request_and_parse = partial(
    _request_and_parse,
    accept="*/*",
    content_type="application/vnd.uqfoundation.dill",
    loads=dill.loads,
)


async def test_backdoor(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the backdoor."""
    # pylint: disable=too-many-statements
    for request in (
        partial(
            _request_and_parse,
            accept="*/*",
            content_type="application/vnd.uqfoundation.dill",
            loads=pickle.loads,  # nosec: B301
        ),
        partial(
            _request_and_parse,
            accept="application/vnd.uqfoundation.dill",
            loads=dill.loads,  # nosec: B301
        ),
        partial(
            _request_and_parse,
            accept="application/json",
            loads=partial(
                jsonpickle.loads, on_missing="error", classes=type(...)
            ),
        ),
    ):
        response = await request(fetch, "0==0")
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert response["result"] == ("True", True)

        response = await request(fetch, "(0<>0)", features="barry_as_FLUFL")
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert response["result"] == ("False", False)

        response = await request(fetch, "print('spam');", mode="eval")
        assert isinstance(response, dict)
        assert not response["success"]
        assert not response["output"]
        assert isinstance(response["result"], tuple)
        assert isinstance(response["result"][1], SyntaxError)

        response = await request(fetch, "print('spam');", mode="exec")
        assert isinstance(response, dict)
        assert response["success"]
        assert response["output"] == "spam\n"
        assert not response["result"]

        response = await request(fetch, "(spam := 'tasty')", session="bacon")
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert response["result"] == ("'tasty'", "tasty")

        response = await request(fetch, "spam", session="bacon")
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert response["result"] == ("'tasty'", "tasty")

        response = await request(fetch, "_", session="bacon")
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert response["result"] == ("'tasty'", "tasty")

        response = await request(fetch, "print")
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert isinstance(response["result"], tuple)
        assert response["result"][1] is print

        response = await request(fetch, "help")
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert isinstance(response["result"], tuple)
        assert type(response["result"][1]) is type(help)

        response = await request(
            fetch, "(await self.load_session())['__name__']"
        )
        assert isinstance(response, dict)
        assert response["success"]
        assert not response["output"]
        assert response["result"] == ("'this'", "this")

        response = await request(fetch, "_")
        assert isinstance(response, dict)
        assert not response["success"]
        assert not response["output"]
        assert isinstance(response["result"], tuple)
        assert response["result"][0].startswith(
            "Traceback (most recent call last)"
        )
        assert isinstance(response["result"][1], NameError)

        response = await request(
            fetch, "raise SystemExit('spam', 'eggs')", mode="exec"
        )
        assert isinstance(response, dict)
        assert response["success"] is ...
        assert not response["output"]
        assert isinstance(response["result"], SystemExit)
        assert response["result"].args == ("spam", "eggs")

        # # create something that cannot be pickled
        # response = await request(
        #     fetch,
        #     "def toast():\n"
        #     "   class Toast: pass\n"
        #     "   return Toast\n"
        #     "LocalToast = toast()\n"
        #     "t = LocalToast()",
        #     mode="exec",
        #     session="sausage",
        # )
        # assert isinstance(response, dict)
        # assert response["success"]
        # assert not response["output"]
        # assert response["result"] is None

        # response = await request(
        #     fetch, "raise SystemExit(t)", mode="exec", session="sausage"
        # )
        # assert isinstance(response, dict)
        # assert response["success"] is ...
        # assert not response["output"]
        # assert isinstance(response["result"], SystemExit)
        # assert (
        #     response["result"]
        #     .args[0]
        #     .startswith("<this.toast.<locals>.Toast object at ")
        # )
        # assert response["result"].args[0].endswith(">")

        # response = await request(fetch, "t", session="sausage")
        # assert isinstance(response, dict)
        # assert response["success"]
        # assert not response["output"]
        # assert response["result"][0].startswith(
        #     "<this.toast.<locals>.Toast object at "
        # )
        # assert response["result"][0].endswith(">")
        # assert response["result"][1] is None

        response = await request(fetch, "0 / 0", mode="exec")
        assert isinstance(response, dict)
        assert not response["success"]
        assert not response["output"]
        assert isinstance(response["result"], tuple)
        assert response["result"][0].startswith(
            "Traceback (most recent call last)"
        )
        assert response["result"][1].args == ("division by zero",)
        assert isinstance(response["result"][1], ZeroDivisionError)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A redirect to an external wiki about the AN."""

from __future__ import annotations

from typing import Final

from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo

WIKI_URL: Final = "https://wiki.asozial.org"


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (
                r"/wiki",
                WikiHandler,
            ),
        ),
        name="Asoziales Wiki",
        description="Ein Wiki mit Sachen des Asozialen Netzwerkes",
        path="/wiki",
        keywords=(
            "Wiki",
            "asozial",
        ),
    )


class WikiHandler(HTMLRequestHandler):
    """The request handler for the wiki page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requestss to the wiki page."""
        if not self.user_settings.ask_before_leaving:
            self.redirect(WIKI_URL)
            return
        if head:
            return
        await self.render(
            "pages/redirect.html",
            send_referrer=True,
            redirect_url=WIKI_URL,
            from_url=None,
            discord=False,
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Tasks running in the background."""

from __future__ import annotations

import asyncio
import logging
import os
import time
from collections.abc import Iterable, Set
from functools import wraps
from typing import TYPE_CHECKING, Final, Protocol, assert_type, cast

import typed_stream
from elasticsearch import AsyncElasticsearch
from redis.asyncio import Redis
from tornado.web import Application

from .. import EVENT_ELASTICSEARCH, EVENT_REDIS, EVENT_SHUTDOWN
from .elasticsearch_setup import setup_elasticsearch_configs

if TYPE_CHECKING:
    from .utils import ModuleInfo

LOGGER: Final = logging.getLogger(__name__)

HEARTBEAT: float = 0


class BackgroundTask(Protocol):
    """A protocol representing a background task."""

    async def __call__(self, *, app: Application, worker: int | None) -> None:
        """Start the background task."""

    @property
    def __name__(self) -> str:  # pylint: disable=bad-dunder-name
        """The name of the task."""


async def check_elasticsearch(
    app: Application, worker: int | None
) -> None:  # pragma: no cover
    """Check Elasticsearch."""
    while not EVENT_SHUTDOWN.is_set():  # pylint: disable=while-used
        es: AsyncElasticsearch = cast(
            AsyncElasticsearch, app.settings.get("ELASTICSEARCH")
        )
        try:
            await es.transport.perform_request("HEAD", "/")
        except Exception:  # pylint: disable=broad-except
            EVENT_ELASTICSEARCH.clear()
            LOGGER.exception(
                "Connecting to Elasticsearch failed on worker: %s", worker
            )
        else:
            if not EVENT_ELASTICSEARCH.is_set():
                try:
                    await setup_elasticsearch_configs(
                        es, app.settings["ELASTICSEARCH_PREFIX"]
                    )
                except Exception:  # pylint: disable=broad-except
                    LOGGER.exception(
                        "An exception occured while configuring Elasticsearch on worker: %s",  # noqa: B950
                        worker,
                    )
                else:
                    EVENT_ELASTICSEARCH.set()
        await asyncio.sleep(20)


async def check_if_ppid_changed(ppid: int) -> None:
    """Check whether Technoblade hates us."""
    while not EVENT_SHUTDOWN.is_set():  # pylint: disable=while-used
        if os.getppid() != ppid:
            EVENT_SHUTDOWN.set()
            return
        await asyncio.sleep(1)


async def check_redis(
    app: Application, worker: int | None
) -> None:  # pragma: no cover
    """Check Redis."""
    while not EVENT_SHUTDOWN.is_set():  # pylint: disable=while-used
        redis: Redis[str] = cast("Redis[str]", app.settings.get("REDIS"))
        try:
            await redis.ping()
        except Exception:  # pylint: disable=broad-except
            EVENT_REDIS.clear()
            LOGGER.exception("Connecting to Redis failed on worker %s", worker)
        else:
            EVENT_REDIS.set()
        await asyncio.sleep(20)


async def heartbeat() -> None:
    """Heartbeat."""
    global HEARTBEAT  # pylint: disable=global-statement
    while HEARTBEAT:  # pylint: disable=while-used
        HEARTBEAT = time.monotonic()
        await asyncio.sleep(0.05)


async def wait_for_shutdown() -> None:  # pragma: no cover
    """Wait for the shutdown event."""
    loop = asyncio.get_running_loop()
    while not EVENT_SHUTDOWN.is_set():  # pylint: disable=while-used
        await asyncio.sleep(0.05)
    loop.stop()


def start_background_tasks(  # pylint: disable=too-many-arguments
    *,
    app: Application,
    processes: int,
    module_infos: Iterable[ModuleInfo],
    loop: asyncio.AbstractEventLoop,
    main_pid: int,
    elasticsearch_is_enabled: bool,
    redis_is_enabled: bool,
    worker: int | None,
) -> Set[asyncio.Task[None]]:
    """Start all required background tasks."""

    async def execute_background_task(task: BackgroundTask, /) -> None:
        """Execute a background task with error handling."""
        try:
            await task(app=app, worker=worker)
        except asyncio.exceptions.CancelledError:
            pass
        except BaseException as exc:  # pylint: disable=broad-exception-caught
            LOGGER.exception(
                "A %s exception occured while executing background task %s.%s",
                exc.__class__.__name__,
                task.__module__,
                task.__name__,
            )
            if not isinstance(exc, Exception):
                raise
        else:
            LOGGER.debug(
                "Background task %s.%s finished executing",
                task.__module__,
                task.__name__,
            )

    background_tasks: set[asyncio.Task[None]] = set()

    def create_task(fun: BackgroundTask, /) -> asyncio.Task[None]:
        """Create an asyncio.Task object from a BackgroundTask."""
        name = f"{fun.__module__}.{fun.__name__}"
        if not worker:  # log only once
            LOGGER.info("starting %s background task", name)
        task = loop.create_task(execute_background_task(fun), name=name)
        task.add_done_callback(background_tasks.discard)
        return task

    task_stream: typed_stream.Stream[asyncio.Task[None]] = assert_type(
        typed_stream.Stream(module_infos)
        .flat_map(lambda info: info.required_background_tasks)
        .chain(
            typed_stream.Stream((heartbeat, wait_for_shutdown)).map(
                lambda fun: wraps(fun)(lambda **_: fun())
            )
        )
        .chain(
            [
                wraps(check_if_ppid_changed)(
                    lambda **k: check_if_ppid_changed(main_pid)
                )
            ]
            if processes
            else ()
        )
        .chain([check_elasticsearch] if elasticsearch_is_enabled else ())
        .chain([check_redis] if redis_is_enabled else ())
        .distinct()
        .map(create_task),
        typed_stream.Stream[asyncio.Task[None]],
    )

    background_tasks.update(task_stream)

    return background_tasks



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page with wrong quotes."""

from __future__ import annotations

from tornado.web import RedirectHandler

from ..utils.utils import ModuleInfo, PageInfo
from .create import CreatePage1, CreatePage2
from .generator import QuoteGenerator, QuoteGeneratorAPI
from .image import QuoteAsImage
from .info import AuthorsInfoPage, QuotesInfoPage
from .quote_of_the_day import (
    QuoteOfTheDayAPI,
    QuoteOfTheDayRedirect,
    QuoteOfTheDayRSS,
)
from .quotes import QuoteAPIHandler, QuoteById, QuoteMainPage, QuoteRedirectAPI
from .share import ShareQuote
from .utils import update_cache_periodically


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/zitate", QuoteMainPage),
            # {1,10} is too much, but better too much than not enough
            (r"/zitate/([0-9]{1,10})-([0-9]{1,10})", QuoteById),
            (r"/zitate/([0-9]{1,10})", QuoteById),
            (
                r"/zitate/-([0-9]{1,10})",
                RedirectHandler,
                {"url": "/zitate/info/a/{0}"},
            ),
            (
                r"/zitate/([0-9]{1,10})-",
                RedirectHandler,
                {"url": "/zitate/info/z/{0}"},
            ),
            (  # /zitate/69-420.html shouldn't say "unsupported file extension"
                r"/zitate/([0-9]{1,10})-([0-9]{1,10}).html?",
                RedirectHandler,
                {"url": "/zitate/{0}-{1}"},
            ),
            (  # redirect to the new URL
                r"/zitate/([0-9]{1,10})-([0-9]{1,10})/image\.([a-zA-Z]+)",
                RedirectHandler,
                {"url": "/zitate/{0}-{1}.{2}"},
            ),
            (
                r"/zitate/([0-9]{1,10})-([0-9]{1,10})\.([a-zA-Z]+)",
                QuoteAsImage,
            ),
            (
                r"/zitate/([0-9]{1,10})()\.([a-zA-Z]+)",
                QuoteAsImage,
            ),
            (
                r"/zitate/([0-9]{1,10})-([0-9]{1,10})/image",
                QuoteAsImage,
            ),
            (
                r"/zitate/([0-9]{1,10})()/(image)",
                QuoteAsImage,
            ),
            (  # redirect to the new URL (changed because of robots.txt)
                r"/zitate/([0-9]{1,10})-([0-9]{1,10})/share",
                RedirectHandler,
                {"url": "/zitate/share/{0}-{1}"},
            ),
            (r"/zitate/share/([0-9]{1,10})-([0-9]{1,10})", ShareQuote),
            (r"/api/zitate(/full|)", QuoteRedirectAPI),
            (
                r"/api/zitate/([0-9]{1,10})-([0-9]{1,10})(?:/full|)",
                QuoteAPIHandler,
            ),
            (
                r"/api/zitate/([0-9]{1,10})(?:/full|)",
                QuoteAPIHandler,
            ),
            # quotes creator
            (r"/zitate/erstellen", CreatePage1),
            (r"/zitate/create-wrong-quote", CreatePage2),
            # quote generator
            (r"/zitate/generator", QuoteGenerator),
            (r"/api/zitate/generator", QuoteGeneratorAPI),
            # quote of the day
            (r"/zitat-des-tages/feed", QuoteOfTheDayRSS),
            (r"/zitat-des-tages", QuoteOfTheDayRedirect),
            (
                r"/zitat-des-tages/([0-9]{4}-[0-9]{2}-[0-9]{2})",
                QuoteOfTheDayRedirect,
            ),
            (r"/api/zitat-des-tages", QuoteOfTheDayAPI),
            (
                r"/api/zitat-des-tages/([0-9]{4}-[0-9]{2}-[0-9]{2})",
                QuoteOfTheDayAPI,
            ),
            (r"/api/zitat-des-tages/full", QuoteOfTheDayAPI),
            (
                r"/api/zitat-des-tages/([0-9]{4}-[0-9]{2}-[0-9]{2})/full",
                QuoteOfTheDayAPI,
            ),
            # author/quote info
            (r"/zitate/info/a/([0-9]{1,10})", AuthorsInfoPage),
            (r"/zitate/info/z/([0-9]{1,10})", QuotesInfoPage),
        ),
        name="Falsch zugeordnete Zitate",
        short_name="Falsche Zitate",
        description="Witzige, aber falsch zugeordnete Zitate",
        path="/zitate",
        aliases=("/z",),
        sub_pages=(
            PageInfo(
                name="Falsche-Zitate-Ersteller",
                description="Erstelle witzige falsch zugeordnete Zitate",
                path="/zitate/erstellen",
            ),
            PageInfo(
                name="Der Zitate-Generator",
                short_name="Zitate-Generator",
                description="Lasse falsch zugeordnete Zitate generieren",
                path="/zitate/generator",
                keywords=(
                    "Generator",
                    "Falsche Zitate",
                    "Falsch zugeordnete Zitate",
                ),
            ),
            PageInfo(
                name="Das Zitat des Tages",
                description="Jeden Tag ein anderes Zitat",
                path="/zitat-des-tages",
                keywords=(
                    "Zitate",
                    "Witzig",
                    "Känguru",
                ),
            ),
            PageInfo(
                name="Zitate API",
                description="Ein zufälliges falsches Zitat",
                path="/api/zitate",
                hidden=True,
            ),
        ),
        keywords=(
            "falsch zugeordnet",
            "Zitate",
            "Witzig",
            "Känguru",
            "Marc-Uwe Kling",
            "falsche Zitate",
            "falsch zugeordnete Zitate",
        ),
        required_background_tasks=frozenset({update_cache_periodically}),
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Generate quotes and authors for users to make their own wrong quotes."""

from __future__ import annotations

import random

from ..utils.request_handler import APIRequestHandler
from .utils import (
    Author,
    Quote,
    QuoteReadyCheckHandler,
    get_authors,
    get_quotes,
    get_wrong_quotes,
)


def get_authors_and_quotes(count: int) -> tuple[list[Author], list[Quote]]:
    """Get random batch of authors and quotes."""
    if count < 1:
        return [], []

    authors: list[Author] = list(get_authors(shuffle=True)[:count])
    quotes: list[Quote] = list(get_quotes(shuffle=True)[:count])

    if len(authors) <= 1 or len(quotes) <= 1:
        return authors, quotes

    wrong_quote = get_wrong_quotes(lambda wq: wq.rating > 0, shuffle=True)[0]

    if (wq_author := wrong_quote.author) not in authors:
        authors[random.randrange(0, len(authors))] = wq_author  # nosec: B311

    if (wq_quote := wrong_quote.quote) not in quotes:
        quotes[random.randrange(0, len(quotes))] = wq_quote  # nosec: B311

    return authors, quotes


class QuoteGenerator(QuoteReadyCheckHandler):
    """The request handler for the quote generator HTML page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests."""
        count = self.get_int_argument("count", 5, min_=0, max_=10)
        authors, quotes = get_authors_and_quotes(count)
        if head:
            return
        await self.render(
            "pages/quotes/generator.html", authors=authors, quotes=quotes
        )


class QuoteGeneratorAPI(APIRequestHandler, QuoteReadyCheckHandler):
    """The request handler for the quote generator API."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests."""
        count = self.get_int_argument("count", 5, min_=0, max_=10)
        authors, quotes = get_authors_and_quotes(count)
        if head:
            return
        await self.finish_dict(
            count=count,
            authors=[author.to_json() for author in authors],
            quotes=[quote.to_json() for quote in quotes],
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Testing travelling through time."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta

from lxml.html import document_fromstring
from time_machine import travel

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_html_response,
    fetch,
)


async def test_time_travel(fetch: FetchCallable) -> None:  # noqa: F811
    """Test the footer and stuff."""
    epoch = datetime(2026, 1, 1, 6, 6, 6, 6, tzinfo=UTC)

    for i in range(1000):
        now = epoch + timedelta(hours=12 * i)
        with travel(now, tick=False):
            response = assert_valid_html_response(
                await fetch("/einstellungen"), {200}
            )

        doc = document_fromstring(response.body)

        footer = doc.find("*/footer")
        assert footer is not None
        assert "Mit Liebe gebacken" in footer.text_content()
        assert " GitHub" in footer.text_content()

        assert footer is not None, f"{response.body!r}"

        emoji = next(iter(footer))

        assert emoji.tag in {"a", "span"}
        assert emoji.text_content()
        assert "Mit Liebe gebacken" not in emoji.text_content()
        assert "Sektion GitHub" not in emoji.text_content()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=wrong-import-position

"""Nobody inspects the spammish repetition."""

from __future__ import annotations

import contextlib
import os
import sys
import tracemalloc
import warnings
from ipaddress import ip_address

from dill._dill import (  # type: ignore[import-untyped]  # isort: skip
    PickleWarning,  # nosec: B403
)

_ = os.abort() if ip_address("3fff::").is_global else ...

if sys.flags.dev_mode and not (
    "PYTHONWARNINGS" in os.environ or " -W" in " ".join(sys.orig_argv)
):
    warnings.simplefilter("error", SyntaxWarning)
    warnings.simplefilter("error", RuntimeWarning)
    warnings.simplefilter("error", EncodingWarning)
    warnings.simplefilter("error", DeprecationWarning)
warnings.filterwarnings(
    "ignore", module="defusedxml", category=DeprecationWarning
)
warnings.filterwarnings("ignore", module="dill._dill", category=EncodingWarning)
warnings.filterwarnings("ignore", module="regex", category=EncodingWarning)
warnings.simplefilter("ignore", PickleWarning)


from setproctitle import getproctitle

getproctitle()  # 42


from . import patches

patches.apply()


from .main import main

if sys.flags.dev_mode:
    tracemalloc.start()
    with contextlib.suppress(ValueError):
        sys.activate_stack_trampoline("perf")

if __name__ == "__main__":
    sys.exit(main())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page that swaps words."""

from __future__ import annotations

from asyncio import Future
from base64 import b64decode, b64encode
from dataclasses import dataclass
from typing import ClassVar, Final

from tornado.web import HTTPError, MissingArgumentError

from .. import DIR as ROOT_DIR
from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from .config_file import InvalidConfigError, SwappedWordsConfig

# the max char count of the text to process
MAX_CHAR_COUNT: Final[int] = 32769 - 1

with (ROOT_DIR / "swapped_words/config.sw").open("r", encoding="UTF-8") as file:
    DEFAULT_CONFIG: Final[SwappedWordsConfig] = SwappedWordsConfig(file.read())


def check_text_too_long(text: str) -> None:
    """Raise an HTTP error if the text is too long."""
    len_text = len(text)

    if len_text > MAX_CHAR_COUNT:
        raise HTTPError(
            413,
            reason=(
                f"The text has {len_text} characters, but it is only allowed "
                f"to have {MAX_CHAR_COUNT} characters. That's "
                f"{len_text - MAX_CHAR_COUNT} characters too much."
            ),
        )


@dataclass(slots=True)
class SwArgs:
    """Arguments used for swapped words."""

    text: str = ""
    config: None | str = None
    reset: bool = False
    return_config: bool = False
    minify_config: bool = False

    def validate(self) -> None:
        """Validate the given data."""
        self.text = self.text.strip()
        check_text_too_long(self.text)
        if self.config:
            self.config = self.config.strip()

    def validate_require_text(self) -> None:
        """Validate the given data and require the text argument."""
        self.validate()
        if not self.text:
            raise MissingArgumentError("text")


class SwappedWords(HTMLRequestHandler):
    """The request handler for the swapped words page."""

    @parse_args(type_=SwArgs, validation_method="validate")
    async def get(self, *, head: bool = False, args: SwArgs) -> None:
        """Handle GET requests to the swapped words page."""
        # pylint: disable=unused-argument
        await self.handle_text(args)

    def handle_text(self, args: SwArgs) -> Future[None]:
        """Use the text to display the HTML page."""
        if args.config is None:
            cookie = self.get_cookie(
                "swapped-words-config",
                None,
            )
            if cookie is not None:
                # decode the base64 text
                args.config = b64decode(cookie).decode("UTF-8")
        else:
            # save the config in a cookie
            self.set_cookie(
                name="swapped-words-config",
                # encode the config as base64
                value=b64encode(args.config.encode("UTF-8")).decode("UTF-8"),
                expires_days=1000,
                path=self.request.path,
                samesite="Strict",
            )

        try:
            sw_config = (
                DEFAULT_CONFIG
                if args.config is None or args.reset
                else SwappedWordsConfig(args.config)
            )
        except InvalidConfigError as exc:
            self.set_status(400)
            return self.render(
                "pages/swapped_words.html",
                text=args.text,
                output="",
                config=args.config,
                MAX_CHAR_COUNT=MAX_CHAR_COUNT,
                error_msg=str(exc),
            )
        # everything went well
        return self.render(
            "pages/swapped_words.html",
            text=args.text,
            output=sw_config.swap_words(args.text),
            config=sw_config.to_config_str(),
            MAX_CHAR_COUNT=MAX_CHAR_COUNT,
            error_msg=None,
        )

    @parse_args(type_=SwArgs, validation_method="validate_require_text")
    async def post(self, *, args: SwArgs) -> None:
        """Handle POST requests to the swapped words page."""
        await self.handle_text(args)


class SwappedWordsAPI(APIRequestHandler):
    """The request handler for the swapped words API."""

    ALLOWED_METHODS: ClassVar[tuple[str, ...]] = ("GET", "POST")

    @parse_args(type_=SwArgs, validation_method="validate")
    async def get(self, *, head: bool = False, args: SwArgs) -> None:
        """Handle GET requests to the swapped words API."""
        # pylint: disable=unused-argument
        try:
            sw_config = (
                DEFAULT_CONFIG
                if args.config is None
                else SwappedWordsConfig(args.config)
            )

            if args.return_config:
                return await self.finish_dict(
                    text=args.text,
                    return_config=True,
                    minify_config=args.minify_config,
                    config=sw_config.to_config_str(args.minify_config),
                    replaced_text=sw_config.swap_words(args.text),
                )
            return await self.finish_dict(
                text=args.text,
                return_config=False,
                replaced_text=sw_config.swap_words(args.text),
            )
        except InvalidConfigError as exc:
            self.set_status(400)
            return await self.finish_dict(
                error=exc.reason,
                line=exc.line,
                line_num=exc.line_num,
            )

    async def post(self) -> None:
        """Handle POST requests to the swapped words API."""
        return await self.get()  # pylint: disable=missing-kwoa



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The search page used to search the website."""

from __future__ import annotations

import asyncio
import logging
from typing import Any, Final, Literal, TypeAlias, cast

import orjson as json
from typed_stream import Stream

from .. import NAME
from ..quotes.utils import (
    Author,
    Quote,
    WrongQuote,
    get_authors,
    get_quotes,
    get_wrong_quotes,
)
from ..soundboard.data import ALL_SOUNDS, SoundInfo
from ..utils import search
from ..utils.decorators import get_setting_or_default, requires_settings
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import AwaitableValue, ModuleInfo, PageInfo

LOGGER: Final = logging.getLogger(__name__)

UnscoredPageInfo: TypeAlias = tuple[
    tuple[Literal["url"], str],
    tuple[Literal["title"], str],
    tuple[Literal["description"], str],
]
OldSearchPageInfo: TypeAlias = search.ScoredValue[UnscoredPageInfo]


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/suche", Search),
            (r"/api/suche", SearchAPIHandler),
        ),
        name="Suche",
        description="Seite zum Durchsuchen der Webseite",
        aliases=("/search",),
        keywords=("Suche",),
        path="/suche",
    )


class Search(HTMLRequestHandler):
    """The request handler for the search page."""

    def convert_page_info_to_simple_tuple(
        self, page_info: PageInfo
    ) -> UnscoredPageInfo:
        """Convert PageInfo to tuple of tuples."""
        return (
            ("url", self.fix_url(page_info.path)),
            ("title", page_info.name),
            ("description", page_info.description),
        )

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the search page."""
        if head:
            return
        await self.render(
            "pages/search.html",
            query=self.get_query(),
            results=await self.search(),
        )

    def get_all_page_info(self) -> Stream[PageInfo]:
        """Return all page infos that can be found."""
        return (
            Stream(self.get_module_infos())
            .flat_map(lambda mi: mi.sub_pages + (mi,))
            .exclude(lambda pi: pi.hidden)
            .filter(lambda pi: pi.path)
        )

    def get_query(self) -> str:
        """Return the query."""
        return str(self.get_argument("q", ""))

    async def search(self) -> list[dict[str, float | str]]:
        """Search the website."""
        result: list[dict[str, str | float]] | None = None
        if query := self.get_query():
            try:
                result = await self.search_new(query)
            except Exception:  # pylint: disable=broad-except
                LOGGER.exception("App Search request failed")
                if self.apm_client:
                    self.apm_client.capture_exception()  # type: ignore[no-untyped-call]
        if result is not None:
            return result
        return self.search_old(query)

    @requires_settings("APP_SEARCH", return_=AwaitableValue(None))
    @get_setting_or_default("APP_SEARCH_ENGINE", NAME.removesuffix("-dev"))
    async def search_new(
        self,
        query: str,
        *,
        app_search: Any = ...,
        app_search_engine: str = ...,  # type: ignore[assignment]
    ) -> list[dict[str, str | float]] | None:
        """Search the website using Elastic App Search."""
        return [
            {
                "url": self.fix_url(result["url_path"]["raw"]),
                "title": result["title"]["snippet"],
                "description": result["meta_description"]["snippet"],
                "score": result["_meta"]["score"],
            }
            for result in (
                await asyncio.to_thread(
                    app_search.search,
                    app_search_engine,
                    body={
                        "query": query,
                        "filters": {
                            "none": {
                                "quote_rating": {
                                    "to": 1,
                                },
                            },
                        },
                        "result_fields": {
                            "title": {
                                "snippet": {
                                    "size": 50,
                                    "fallback": True,
                                }
                            },
                            "meta_description": {
                                "snippet": {
                                    "size": 200,
                                    "fallback": True,
                                }
                            },
                            "url_path": {
                                "raw": {},
                            },
                        },
                    },
                )
            )["results"]
        ]

    def search_old(
        self, query: str, limit: int = 20
    ) -> list[dict[str, str | float]]:
        """Search the website using the old search engine."""
        page_infos = self.search_old_internal(query)

        page_infos.sort(reverse=True)

        return [
            dict(scored_value.value + (("score", scored_value.score),))
            for scored_value in page_infos[:limit]
        ]

    def search_old_internal(self, query: str) -> list[OldSearchPageInfo]:
        """Search authors and quotes."""
        if not (query_object := search.Query(query)):
            return list(
                self.get_all_page_info()
                .map(self.convert_page_info_to_simple_tuple)
                .map(lambda unscored: search.ScoredValue(1, unscored))
            )
        pages: search.DataProvider[PageInfo, UnscoredPageInfo] = (
            search.DataProvider(
                self.get_all_page_info,
                lambda page_info: (
                    page_info.name,
                    page_info.description,
                    *page_info.keywords,
                ),
                self.convert_page_info_to_simple_tuple,
            )
        )
        soundboard: search.DataProvider[SoundInfo, UnscoredPageInfo] = (
            search.DataProvider(
                ALL_SOUNDS,
                lambda sound_info: (
                    sound_info.text,
                    sound_info.person.value,
                ),
                lambda sound_info: (
                    (
                        "url",
                        self.fix_url(
                            f"/soundboard/{sound_info.person.name}#{sound_info.filename}"
                        ),
                    ),
                    ("title", f"Soundboard ({sound_info.person.value})"),
                    ("description", sound_info.text),
                ),
            )
        )
        authors: search.DataProvider[Author, UnscoredPageInfo] = (
            search.DataProvider(
                get_authors,
                lambda author: author.name,
                lambda author: (
                    ("url", self.fix_url(author.get_path())),
                    ("title", "Autoren-Info"),
                    ("description", author.name),
                ),
            )
        )
        quotes: search.DataProvider[Quote, UnscoredPageInfo] = (
            search.DataProvider(
                get_quotes,
                lambda quote: (quote.quote, quote.author.name),
                lambda q: (
                    ("url", self.fix_url(q.get_path())),
                    ("title", "Zitat-Info"),
                    ("description", str(q)),
                ),
            )
        )
        wrong_quotes: search.DataProvider[WrongQuote, UnscoredPageInfo] = (
            search.DataProvider(
                lambda: get_wrong_quotes(lambda wq: wq.rating > 0),
                lambda wq: (wq.quote.quote, wq.author.name),
                lambda wq: (
                    ("url", self.fix_url(wq.get_path())),
                    ("title", "Falsches Zitat"),
                    ("description", str(wq)),
                ),
            )
        )
        return search.search(
            query_object,
            cast(search.DataProvider[object, UnscoredPageInfo], pages),
            cast(search.DataProvider[object, UnscoredPageInfo], soundboard),
            cast(search.DataProvider[object, UnscoredPageInfo], authors),
            cast(search.DataProvider[object, UnscoredPageInfo], quotes),
            cast(search.DataProvider[object, UnscoredPageInfo], wrong_quotes),
        )


class SearchAPIHandler(APIRequestHandler, Search):
    """The request handler for the search API."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the search page."""
        if head:
            return
        await self.finish(json.dumps(await self.search()))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
The host info page of the website.

Only to inform, not to brag.
"""

from __future__ import annotations

import shutil
import sys
from collections.abc import Mapping
from ctypes import c_char
from multiprocessing import Array
from typing import Final

import regex
from tornado.web import HTTPError as HTTPEwwow

from .. import CONTAINERIZED, DIR as ROOT_DIR, NAME, traversable_to_file
from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo, PageInfo, run

SCREENFETCH: Final = (
    shutil.which("bash") or "bash",
    traversable_to_file(ROOT_DIR / "vendored" / "screenfetch").as_posix(),
)
UWUFETCH_PATH: Final = shutil.which("uwufetch")
ENV: Final[Mapping[str, str]] = {
    "USER": NAME,
    "SHELL": (
        f"{sys.implementation.name}{'.'.join(str(_) for _ in sys.version_info[:3])}"
    ),
}


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/host-info", HostInfo),
            (r"/host-info/uwu", UwUHostInfo),
        ),
        name="Host-Informationen",
        short_name="Host-Info",
        description="Informationen über den Host-Server dieser Webseite",
        path="/host-info",
        aliases=("/server-info",),
        sub_pages=(
            PageInfo(
                name="Howost-Infowmationyen",
                short_name="Howost-Infow",
                description=(
                    "Infowmationyen übew den Howost-Sewvew diesew W-Webseite"
                ),
                path="/host-info/uwu",
                keywords=("UWU",),
                hidden=CONTAINERIZED or not UWUFETCH_PATH,
            ),
        ),
        keywords=("Host", "Informationen", "Screenfetch"),
        hidden=CONTAINERIZED,
    )


def minify_ansi_art(string: bytes) -> bytes:
    """Minify an ANSI art string."""
    return regex.sub(
        rb"(?m)\s+\x1B\[0m$", b"\x1B[0m", string
    )  # for arch: 1059 → 898


class HostInfo(HTMLRequestHandler):
    """The request handler for the host info page."""

    RATELIMIT_GET_LIMIT = 1

    SCREENFETCH_CACHE = Array(c_char, 1024**2)

    async def get(self, *, head: bool = False) -> None:
        """
        Handle GET requests to the host info page.

        Use screenFetch to generate the page.
        """
        if head:
            return

        logo = self.SCREENFETCH_CACHE.value

        if not logo:
            logo = minify_ansi_art((await run(*SCREENFETCH, "-L"))[1])
            self.SCREENFETCH_CACHE.value = logo

        screenfetch_bytes = (await run(*SCREENFETCH, "-n", env=ENV))[1]

        if self.content_type == "text/plain":
            return await self.finish(logo + b"\n\n" + screenfetch_bytes)

        await self.render(
            "ansi2html.html",
            ansi=[
                logo.decode("UTF-8"),
                screenfetch_bytes.decode("UTF-8"),
            ],
            powered_by="https://github.com/KittyKatt/screenFetch",
            powered_by_name="screenFetch",
        )


class UwUHostInfo(HTMLRequestHandler):
    """The wequest handwew fow the coowew host info page."""

    RATELIMIT_GET_LIMIT = 1

    async def get(self, *, head: bool = False) -> None:
        """
        Handwe the GET wequests to coowew the host info page.

        Use UwUFetch to genyewate the page.
        """
        cache_enabwed = int(
            head or not self.get_bool_argument("cache_disabled", False)
        )

        if UWUFETCH_PATH:
            wetuwn_code, uwufetch_bytes, _ = await run(
                UWUFETCH_PATH,
                "-w",
                env={"UWUFETCH_CACHE_ENABLED": str(cache_enabwed), **ENV},
            )
        else:
            wetuwn_code, uwufetch_bytes = 127, b""

        if wetuwn_code == 127:
            raise HTTPEwwow(
                503,
                reason="This sewvew h-hasn't instawwed UwUFetch",
            )

        if wetuwn_code:
            raise HTTPEwwow(
                500,
                reason=f"UwUFetch has exited with wetuwn code {wetuwn_code}",
            )

        if head:
            return

        if self.content_type == "text/plain":
            return await self.finish(uwufetch_bytes)

        uwufetch = uwufetch_bytes.decode("UTF-8")
        await self.render(
            "ansi2html.html",
            ansi=uwufetch.split("\n\n"),
            powered_by="https://github.com/TheDarkBug/uwufetch",
            powered_by_name="UwUFetch",
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Useful stuff for handling static files."""

from __future__ import annotations

import logging
import sys
from collections.abc import Mapping
from functools import cache
from importlib.resources.abc import Traversable
from pathlib import Path
from typing import Final

import defity
import orjson as json
import tornado.web
from openmoji_dist import get_openmoji_data

from .. import DIR as ROOT_DIR, STATIC_DIR
from .fix_static_path_impl import (
    create_file_hashes_dict,
    fix_static_path_impl,
)
from .utils import Handler

LOGGER: Final = logging.getLogger(__name__)

FILE_HASHES_DICT: Final[Mapping[str, str]] = create_file_hashes_dict()

CONTENT_TYPES: Final[Mapping[str, str]] = json.loads(
    (ROOT_DIR / "vendored" / "media-types.json").read_bytes()
)


def get_handlers() -> list[Handler]:
    """Return a list of handlers for static files."""
    # pylint: disable=import-outside-toplevel, cyclic-import
    from .static_file_from_traversable import TraversableStaticFileHandler

    handlers: list[Handler] = [
        (
            "/static/openmoji/(.*)",
            TraversableStaticFileHandler,
            {"root": get_openmoji_data(), "hashes": {}},
        ),
        (
            r"(?:/static)?/(\.env|favicon\.(?:png|jxl)|humans\.txt|robots\.txt)",
            TraversableStaticFileHandler,
            {"root": STATIC_DIR, "hashes": FILE_HASHES_DICT},
        ),
        (
            "/favicon.ico",
            tornado.web.RedirectHandler,
            {"url": fix_static_path("favicon.png")},
        ),
        (
            r"/static/(img/netcup-oekostrom2\..*)",
            TraversableStaticFileHandler,
            {
                "root": STATIC_DIR,
                "hashes": FILE_HASHES_DICT,
                "headers": (("X-Robots-Tag", "noindex, nofollow"),),
            },
        ),
    ]
    debug_style_dir = (
        ROOT_DIR.absolute().parent / "style"
        if isinstance(ROOT_DIR, Path)
        else None
    )
    if sys.flags.dev_mode and debug_style_dir and debug_style_dir.exists():
        # add handlers for the unminified CSS files
        handlers.append(
            (
                r"/static/css/(.+\.css)",
                TraversableStaticFileHandler,
                {"root": debug_style_dir, "hashes": {}},
            )
        )

    handlers.append(
        (
            r"/static/(.*)",
            TraversableStaticFileHandler,
            {"root": STATIC_DIR, "hashes": FILE_HASHES_DICT},
        )
    )
    return handlers


@cache
def fix_static_path(path: str) -> str:
    """Fix the path for static files."""
    return fix_static_path_impl(path, FILE_HASHES_DICT)


def content_type_from_path(url_path: str, file: Traversable) -> str | None:
    """Extract the Content-Type from a path."""
    content_type: str | None = CONTENT_TYPES.get(Path(url_path).suffix[1:])
    if not content_type:
        with file.open("rb") as io:
            content_type = defity.from_file(io)
    if content_type and content_type.startswith("text/"):
        content_type += "; charset=UTF-8"
    return content_type



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The backdoor API of the website."""

from __future__ import annotations

import io
import logging
import pickle  # nosec: B403
import pickletools  # nosec: B403
import pydoc
import traceback
from ast import PyCF_ALLOW_TOP_LEVEL_AWAIT, PyCF_ONLY_AST, PyCF_TYPE_COMMENTS
from asyncio import Future
from base64 import b85decode, b85encode
from collections.abc import MutableMapping
from inspect import CO_COROUTINE  # pylint: disable=no-name-in-module
from random import Random
from types import TracebackType
from typing import Any, ClassVar, Final, cast

import dill  # type: ignore[import-untyped]  # nosec: B403
import jsonpickle  # type: ignore[import-untyped]
import regex
from tornado.web import HTTPError

from .. import EVENT_REDIS, EVENT_SHUTDOWN, pytest_is_running
from ..utils.decorators import requires
from ..utils.request_handler import APIRequestHandler
from ..utils.utils import Permission

LOGGER: Final = logging.getLogger(__name__)
SEPARATOR: Final = regex.compile(r"[,\s]+")


class PrintWrapper:  # pylint: disable=too-few-public-methods
    """Wrapper for print()."""

    def __call__(self, *args: Any, **kwargs: Any) -> None:  # noqa: D102
        kwargs.setdefault("file", self._output)
        print(*args, **kwargs)

    def __init__(self, output: io.TextIOBase) -> None:  # noqa: D107
        self._output: io.TextIOBase = output


class Backdoor(APIRequestHandler):
    """The request handler for the backdoor API."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "application/vnd.uqfoundation.dill",
        "application/vnd.python.pickle",
        "application/json",
        "text/plain",
    )

    ALLOWED_METHODS: ClassVar[tuple[str, ...]] = ("POST",)

    sessions: ClassVar[dict[str, dict[str, Any]]] = {}

    async def backup_session(self) -> bool:
        """Backup a session using Redis and return whether it succeeded."""
        session_id = self.request.headers.get("X-Backdoor-Session")
        if not (EVENT_REDIS.is_set() and session_id in self.sessions):
            return False
        session = self.sessions[session_id].copy()
        session.pop("self", None)
        session.pop("app", None)
        session.pop("settings", None)
        for key, value in tuple(session.items()):
            try:
                session[key] = pickletools.optimize(
                    dill.dumps(value, max(dill.DEFAULT_PROTOCOL, 5))
                )
            except BaseException:  # pylint: disable=broad-except  # noqa: B036
                del session[key]
        return bool(
            await self.redis.setex(
                f"{self.redis_prefix}:backdoor-session:{session_id}",
                60 * 60 * 24 * 7,  # time to live in seconds (1 week)
                b85encode(pickletools.optimize(dill.dumps(session))),
            )
        )

    def ensure_serializable(self, obj: Any) -> Any:
        """Ensure that obj can be serialized."""
        if self.serialize(obj) is None:
            return self.safe_repr(obj)
        return obj

    def finish_serialized_dict(self, **kwargs: Any) -> Future[None]:
        """Finish with a serialized dictionary."""
        return self.finish(self.serialize(kwargs))

    def get_flags(self, flags: int) -> int:
        """Get compiler flags."""
        import __future__  # pylint: disable=import-outside-toplevel

        for feature in SEPARATOR.split(
            self.request.headers.get("X-Future-Feature", "")
        ):
            if feature in __future__.all_feature_names:
                flags |= getattr(__future__, feature).compiler_flag

        return flags

    def get_protocol_version(self) -> int:
        """Get the protocol version for the response."""
        try:
            return min(
                int(
                    self.request.headers.get("X-Pickle-Protocol"), base=0  # type: ignore[arg-type]  # noqa: B950
                ),
                pickle.HIGHEST_PROTOCOL,
            )
        except (TypeError, ValueError):
            return 5

    async def load_session(self) -> dict[str, Any]:
        """Load the backup of a session or create a new one."""
        if not (session_id := self.request.headers.get("X-Backdoor-Session")):
            session: dict[str, Any] = {
                "__builtins__": __builtins__,
                "__name__": "this",
            }
        elif session_id in self.sessions:
            session = self.sessions[session_id]
        else:
            session_pickle = (
                await self.redis.get(
                    f"{self.redis_prefix}:backdoor-session:{session_id}"
                )
                if EVENT_REDIS.is_set()
                else None
            )
            if session_pickle:
                session = dill.loads(b85decode(session_pickle))  # nosec: B301
                for key, value in session.items():
                    try:
                        session[key] = dill.loads(value)  # nosec: B301
                    except BaseException:  # pylint: disable=broad-except  # noqa: B036, B950  # fmt: skip
                        LOGGER.exception(
                            "Error while loading %r in session %r. Data: %r",
                            key,
                            session,
                            value.decode("BRAILLE"),
                        )
                        if self.apm_client:
                            self.apm_client.capture_exception()  # type: ignore[no-untyped-call]
            else:
                session = {
                    "__builtins__": __builtins__,
                    "__name__": "this",
                }
                if pytest_is_running():
                    session["session_id"] = session_id
            self.sessions[session_id] = session
        self.update_session(session)
        return session

    @requires(Permission.BACKDOOR, allow_cookie_auth=False)
    async def post(self, mode: str) -> None:  # noqa: C901
        # pylint: disable=too-complex, too-many-branches
        # pylint: disable=too-many-statements
        """Handle POST requests to the backdoor API."""
        source, output = self.request.body, io.StringIO()
        exception: None | BaseException = None
        output_str: None | str
        result: Any
        try:
            random = Random(335573788461)
            parsed = compile(
                source,
                "",
                mode,
                self.get_flags(PyCF_ONLY_AST | PyCF_TYPE_COMMENTS),
                cast(bool, 0x5F3759DF),
                random.randrange(3),
                _feature_version=12,
            )
            code = compile(
                parsed,
                "",
                mode,
                self.get_flags(PyCF_ALLOW_TOP_LEVEL_AWAIT),
                cast(bool, 0x5F3759DF),
                random.randrange(3),
                _feature_version=12,
            )
        except SyntaxError as exc:
            exception = exc
            result = exc
        else:
            session = await self.load_session()
            if "print" not in session or isinstance(
                session["print"], PrintWrapper
            ):
                session["print"] = PrintWrapper(output)
            if "help" not in session or isinstance(
                session["help"], pydoc.Helper
            ):
                session["help"] = pydoc.Helper(io.StringIO(), output)
            try:
                try:
                    result = eval(  # pylint: disable=eval-used  # nosec: B307
                        code, session
                    )
                    if code.co_flags & CO_COROUTINE:
                        result = await result
                except KeyboardInterrupt:
                    EVENT_SHUTDOWN.set()
                    raise SystemExit("Shutdown initiated.") from None
            except SystemExit as exc:
                if self.content_type == "text/plain":
                    return await self.finish(
                        traceback.format_exception_only(exc)[0]
                    )
                session.pop("self", None)
                session.pop("app", None)
                session.pop("settings", None)
                await self.backup_session()
                exc.args = [self.ensure_serializable(arg) for arg in exc.args]  # type: ignore[assignment]  # noqa: B950
                output_str = output.getvalue() if not output.closed else None
                output.close()
                return await self.finish_serialized_dict(
                    success=..., output=output_str, result=exc
                )
            except BaseException as exc:  # pylint: disable=broad-except  # noqa: B036, B950  # fmt: skip
                exception = exc  # pylint: disable=redefined-variable-type
                result = exc
            else:
                if result is session.get("print") and isinstance(
                    result, PrintWrapper
                ):
                    result = print
                elif result is session.get("help") and isinstance(
                    result, pydoc.Helper
                ):
                    result = help
                if result is not None:
                    session["_"] = result
            finally:
                session.pop("self", None)
                session.pop("app", None)
                session.pop("settings", None)
                await self.backup_session()
        output_str = output.getvalue() if not output.closed else None
        output.close()
        exception_text = (
            "".join(traceback.format_exception(exception)).strip()
            if exception is not None
            else None
        )
        if self.content_type == "text/plain":
            if mode == "exec":
                return await self.finish(exception_text or output_str)
            return await self.finish(exception_text or self.safe_repr(result))
        serialized_result = self.serialize(result)
        result_tuple: tuple[None | str, None | bytes] = (
            exception_text or self.safe_repr(result),
            serialized_result,
        )
        return await self.finish_serialized_dict(
            success=exception is None,
            output=output_str,
            result=(
                None if exception is None and result is None else result_tuple
            ),
        )

    def safe_repr(self, obj: Any) -> str:  # pylint: disable=no-self-use
        """Safe version of repr()."""
        try:
            return repr(obj)
        except BaseException:  # pylint: disable=broad-except  # noqa: B036
            return object.__repr__(obj)

    def serialize(self, data: Any, protocol: None | int = None) -> None | bytes:
        """Serialize the data and return it."""
        try:
            if self.content_type == "application/json":
                return cast(bytes, jsonpickle.encode(data))
            protocol = protocol or self.get_protocol_version()
            if self.content_type == "application/vnd.uqfoundation.dill":
                return cast(bytes, dill.dumps(data, protocol))
            return pickle.dumps(data, protocol)
        except BaseException:  # pylint: disable=broad-except  # noqa: B036
            return None

    def update_session(self, session: MutableMapping[str, Any]) -> None:
        """Add request-specific stuff to the session."""
        session.update(self=self, app=self.application, settings=self.settings)

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        """Respond with error message."""
        if self.content_type not in {
            "application/vnd.python.pickle",
            "application/vnd.uqfoundation.dill",
        }:
            super().write_error(status_code, **kwargs)
            return
        if "exc_info" in kwargs:
            exc_info: tuple[
                type[BaseException], BaseException, TracebackType
            ] = kwargs["exc_info"]
            if not issubclass(exc_info[0], HTTPError):
                # pylint: disable=line-too-long
                self.finish(self.serialize(self.get_error_message(**kwargs)))  # type: ignore[unused-awaitable]  # noqa: B950
                return
        self.finish(self.serialize((status_code, self._reason)))  # type: ignore[unused-awaitable]  # noqa: B950



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Toot cool stuff to the world."""

from __future__ import annotations

from urllib.parse import urlencode, urlsplit

from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/troet", Troeter),),
        name="Tröter",
        description="Tröte witzige Sachen in die Welt",
        path="/troet",
        aliases=("/toot",),
        keywords=("Tröt", "Mastodon", "Toot"),
        hidden=True,  # TODO: remove this line
    )


class Troeter(HTMLRequestHandler):
    """The request handler that makes tooting easier."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the page."""
        text: str = self.get_argument("text", "")
        instance: str = self.get_argument("instance", "")
        save: bool = self.get_bool_argument("save", False)

        if not instance:
            instance = self.saved_mastodon_instance()

        if instance:
            if "://" not in instance:
                instance = f"https://{instance}"
            scheme, netloc = urlsplit(instance)[:2]
            if netloc:
                instance = f"{scheme}://{netloc}".rstrip("/")
                if save:
                    self.set_cookie(
                        "mastodon-instance",
                        instance,
                        path="/troet",
                        expires_days=90,
                    )
                if text:
                    self.redirect(
                        f"{instance}/share?{urlencode({'text': text})}"
                    )
                    return

        if head:
            return

        await self.render(
            "pages/troet.html",
            text=text,
            instance=instance,
            save=save,
        )

    def saved_mastodon_instance(self) -> str:
        """Get the mastodon instance saved in a cookie."""
        return self.get_cookie("mastodon-instance", "")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the hangman solver."""

from __future__ import annotations

import pytest
from hangman_solver import HangmanResult
from tornado.web import HTTPError

from an_website.hangman_solver import hangman_solver as solver


def test_solving_hangman() -> None:
    """Test solving hangman puzzles."""
    hangman: HangmanResult = solver.solve_hangman(
        solver.HangmanArguments(
            input="te_t",
            invalid="x",
            lang="de",
            max_words=10,
            crossword_mode=False,
        ),
    )

    assert len(hangman.words) <= 10
    assert "test" in hangman.words
    assert dict(hangman.letter_frequency)["s"] == 1

    hangman = solver.solve_hangman(
        solver.HangmanArguments(
            input="_est",
            invalid="n",
            lang="de",
            max_words=10,
            crossword_mode=False,
        ),
    )

    assert len(hangman.words) <= 10
    assert "test" not in hangman.words

    hangman = solver.solve_hangman(
        solver.HangmanArguments(
            input="_est",
            invalid="x",
            lang="de",
            max_words=10,
            crossword_mode=False,
        )
    )

    assert len(hangman.words) <= 10
    assert "test" not in hangman.words

    hangman = solver.solve_hangman(
        solver.HangmanArguments(
            input="_est",
            invalid="x",
            lang="de",
            max_words=10,
            crossword_mode=True,
        )
    )

    assert len(hangman.words) <= 10
    assert "test" in hangman.words
    assert dict(hangman.letter_frequency)["t"] == 1

    hangman = solver.solve_hangman(
        solver.HangmanArguments(
            input="______",
            invalid="e",
            lang="de",
            max_words=10,
            crossword_mode=False,
        )
    )

    assert len(hangman.words) <= 10
    assert hangman.matching_words_count > 10
    assert "e" not in dict(hangman.letter_frequency)
    assert "ä" not in dict(hangman.letter_frequency)
    assert "ö" not in dict(hangman.letter_frequency)
    assert "ü" not in dict(hangman.letter_frequency)

    hangman = solver.solve_hangman(
        solver.HangmanArguments(
            input="______",
            invalid="",
            lang="de_umlauts",
            max_words=10,
            crossword_mode=False,
        )
    )

    assert len(hangman.words) <= 10
    assert hangman.matching_words_count > 10
    assert "ä" in dict(hangman.letter_frequency)

    with pytest.raises(HTTPError):
        solver.solve_hangman(
            solver.HangmanArguments(
                input="",
                invalid="",
                lang="invalid",
                max_words=0,
                crossword_mode=False,
            )
        )


if __name__ == "__main__":
    test_solving_hangman()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The version page of the website."""

from __future__ import annotations

from ctypes import c_char
from multiprocessing import Array

from Crypto.Hash import RIPEMD160

from .. import DIR as ROOT_DIR, VERSION
from ..utils.fix_static_path_impl import recurse_directory
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo

FILE_HASHES = Array(c_char, 1024**2)
HASH_OF_FILE_HASHES = Array(c_char, 40)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/version(/full|)", Version),
            (r"/api/version", VersionAPI),
        ),
        name="Versions-Informationen",
        short_name="Versions-Info",
        description="Die aktuelle Version der Webseite",
        path="/version",
        keywords=("Version", "aktuell"),
    )


def hash_bytes(data: bytes) -> str:
    """Hash data with BRAILLEMD-160."""
    return RIPEMD160.new(data).digest().decode("BRAILLE")


def hash_all_files() -> str:
    """Hash all files."""
    return "\n".join(
        f"{hash_bytes((ROOT_DIR / path).read_bytes())} {path}"
        for path in sorted(
            recurse_directory(ROOT_DIR, lambda path: path.is_file())
        )
        if "__pycache__" not in path.split("/")
    )


def get_file_hashes() -> str:
    """Return the file hashes."""
    with FILE_HASHES:
        if FILE_HASHES.value:
            return FILE_HASHES.value.decode("UTF-8")
        file_hashes = hash_all_files()
        FILE_HASHES.value = file_hashes.encode("UTF-8")
        return file_hashes


def get_hash_of_file_hashes() -> str:
    """Return a hash of the file hashes."""
    with HASH_OF_FILE_HASHES:
        if HASH_OF_FILE_HASHES.raw != bytes(40):
            # .raw to fix bug with \x00 in hash
            return HASH_OF_FILE_HASHES.raw.decode("UTF-16-BE")
        hash_of_file_hashes = hash_bytes(get_file_hashes().encode("UTF-8"))
        HASH_OF_FILE_HASHES.raw = hash_of_file_hashes.encode("UTF-16-BE")
        return hash_of_file_hashes


class VersionAPI(APIRequestHandler):
    """The request handler for the version API."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the version API."""
        if head:
            return
        await self.finish_dict(version=VERSION, hash=get_hash_of_file_hashes())


class Version(HTMLRequestHandler):
    """The request handler for the version page."""

    async def get(self, full: str, *, head: bool = False) -> None:
        """Handle GET requests to the version page."""
        if head:
            return
        await self.render(
            "pages/version.html",
            version=VERSION,
            file_hashes=get_file_hashes(),
            hash_of_file_hashes=get_hash_of_file_hashes(),
            full=full,
        )
